import os
import shutil

import typer

from .preview import preview_cli
from .project import project_cli
from .sources import sources_cli

app = typer.Typer(pretty_exceptions_show_locals=False)
app.add_typer(preview_cli, name="preview")
app.add_typer(sources_cli, name="sources")
app.add_typer(project_cli, name="project")


@app.command("init")
def init_project(project_name: str):
    """Initialize a new Vinyl project"""
    normalized_project_name = project_name.lower().replace(" ", "_")
    scaffolding_path = os.path.join(os.path.dirname(__file__), "_project_scaffolding")
    project_path = os.path.join(os.getcwd(), normalized_project_name)
    if os.path.exists(project_path):
        raise ValueError(f"Directory {project_path} already exists")

    # copy the scaffolding to the new project path
    shutil.copytree(scaffolding_path, project_path)
    # rename the project folder
    os.rename(
        os.path.join(project_path, "__project_name__"),
        os.path.join(project_path, normalized_project_name),
    )

    # templatize project assets
    project_assets = [
        "README.md",
        "pyproject.toml",
    ]
    asset_paths = [os.path.join(project_path, asset) for asset in project_assets]
    for path in asset_paths:
        with open(path, "r") as f:
            content = f.read()
        content = content.replace("{{PROJECT_NAME}}", normalized_project_name)
        with open(path, "w") as f:
            f.write(content)

    typer.echo(f"Created project at {project_path}")


if __name__ == "__main__":
    app()



from ibis import Schema
from rich.table import Table
from rich.text import Text
from textual.app import App
from textual.binding import Binding


def adjust_type(type) -> Text:
    if type.is_numeric():
        return Text(str(type), style="yellow")
    elif type.is_string():
        return Text(str(type), style="orange1")
    elif type.is_temporal():
        return Text(str(type), style="cyan")
    elif type.is_boolean():
        return Text(str(type), style="purple")
    elif type.is_json() or type.is_struct() or type.is_array():
        return Text(str(type), style="bright_magenta")
    elif type.is_uuid():
        return Text(str(type), style="medium_purple1")
    elif type.is_geospatial():
        return Text(str(type), style="dark_goldenrod")
    else:
        return Text(str(type))


def print_schema(schema: Schema):
    out = Table()
    out.add_column("#")
    out.add_column("column")
    out.add_column("type")
    for i, (name, type_) in enumerate(schema.items()):
        out.add_row(str(i), str(name), adjust_type(type_))

    return out


class TurntableTextualApp(App):
    BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
        Binding(key="r", action="reload", description="Reload"),
        Binding(key="d", action="toggle_dark", description="Toggle Dark Mode"),
    ]

    def action_reload(self) -> None:
        self.mount()



from __future__ import annotations

import pickle as pkl
from functools import reduce, wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.types as ir
import ibis.selectors as s
import sqlglot
from ibis import Schema, _
from sqlglot.optimizer import optimize
from sqlglot.optimizer.eliminate_ctes import eliminate_ctes
from sqlglot.optimizer.eliminate_joins import eliminate_joins
from sqlglot.optimizer.eliminate_subqueries import eliminate_subqueries
from sqlglot.optimizer.normalize import normalize
from sqlglot.optimizer.pushdown_predicates import pushdown_predicates
from sqlglot.optimizer.pushdown_projections import pushdown_projections
from sqlglot.optimizer.unnest_subqueries import unnest_subqueries

from vinyl.lib.column import (
    VinylColumn,
    VinylColumnList,
    VinylSortColumnList,
    base_boolean_column_type,
    base_column_type,
    boolean_column_type,
    column_type,
    column_type_all,
    column_type_without_dict,
)
from vinyl.lib.enums import FillOptions, WindowType
from vinyl.lib.graph import unify_backends
from vinyl.lib.schema import VinylSchema
from vinyl.lib.table_methods import (
    _adjust_fill_list,
    _build_spine,
    _difference_two,
    _join_spine,
    _process_multiple_select,
    _union_two,
    fill_type,
)
from vinyl.lib.utils.functions import validate

if TYPE_CHECKING:
    from vinyl.lib.chart import geom

RULES = (
    # qualify,
    pushdown_projections,
    normalize,
    unnest_subqueries,
    pushdown_predicates,
    # optimize_joins,  # produces incorrect results
    eliminate_subqueries,
    # merge_subqueries,  # produces incorrect results
    eliminate_joins,
    eliminate_ctes,
    # quote_identifiers,
    # annotate_types,
    # canonicalize,
    # simplify,
)


def ensure_output(func):
    @wraps(func)
    def wrapper(self: VinylTable, *args, **kwargs) -> VinylTable:
        # cache current objects if immutable
        if not self.mutable:
            cached_arg = self._arg
            cached_conn_replace = self._conn_replace
            cached_twin_conn_replace = self._twin_conn_replace

        # build the connection dict and set an updated version to be used in the function
        combined_conn_replace = {}
        combined_twin_conn_replace = {}
        for el in [self, *args, *list(kwargs.values())]:
            if isinstance(el, VinylTable):
                combined_conn_replace.update(el._conn_replace)
                combined_twin_conn_replace.update(el._twin_conn_replace)

        result = func(self, *args, **kwargs)

        if not self.mutable:
            self._arg = cached_arg
            self._conn_replace = cached_conn_replace
            self._twin_conn_replace = cached_twin_conn_replace
        else:
            self._arg = result._arg
            self.mutable = True  # chain mutability to subsequent function calls
            self._conn_replace = combined_conn_replace
            self._twin_conn_replace = combined_twin_conn_replace

        self.reset_columns()

        final = VinylTable(
            result._arg,
            _conn_replace=combined_conn_replace,
            _twin_conn_replace=combined_twin_conn_replace,
        )
        final.mutable = self.mutable
        return final

    return wrapper


class VinylTable:
    _arg: ir.Expr
    mutable: bool
    _conn_replace: dict[ops.Relation, ops.Relation]
    _twin_conn_replace: dict[ops.Relation, ops.Relation]

    def __init__(
        self,
        _arg: ir.Expr,
        _conn_replace: dict[ops.Relation, ops.Relation] = {},
        _twin_conn_replace: dict[ops.Relation, ops.Relation] = {},
    ):
        self._arg = _arg
        self.mutable = False
        self._conn_replace = _conn_replace
        self._twin_conn_replace = _twin_conn_replace
        self.set_columns()

    def set_columns(self, deferred=False):
        # set columns
        for i, (name, type_) in enumerate(self.tbl.schema().items()):
            col = getattr(_, name) if deferred else getattr(self.tbl, name)
            setattr(self, name, col)
            setattr(self, f"_{i}", col)

    def clear_columns(self):
        for name in self.__dict__.copy():
            if name not in ["_arg", "mutable", "_conn_replace", "_twin_conn_replace"]:
                delattr(self, name)

    def reset_columns(self, deferred=False):
        self.clear_columns()
        self.set_columns(deferred=deferred)

    # Make callable so that @model wrapper works
    def __call__(self):
        return self

    def __enter__(self):
        # Create a copy of the original object and make the object mutable
        new = self.copy()
        new.mutable = True
        return new

    def __exit__(self, exc_type, exc_value, traceback):
        # Exit logic here
        pass

    def __str__(self):
        return self.tbl.__str__()

    def __repr__(self):
        return self.tbl.__repr__()

    @property
    def tbl(self):
        return ir.Table(self._arg)

    def copy(self) -> VinylTable:
        return VinylTable(self._arg, self._conn_replace, self._twin_conn_replace)

    @ensure_output
    def __add__(self, other) -> VinylTable:
        # Handle adding CustomNumber to 0, which is the default start value for sum
        if not isinstance(other, VinylTable):
            if other == 0:
                return self

            raise ValueError(
                f"Can only add a VinylTable to another VinylTable, not a {type(other)}"
            )

        return _union_two(self.tbl, other.tbl)

    def __radd__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only add a VinylTable to another VinylTable, not a {type(other)}"
            )
        return other.__add__(self)

    @ensure_output
    def __sub__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only subtract a VinylTable from another VinylTable, not a {type(other)}"
            )
        return _difference_two(self.tbl, other.tbl)

    def __rsub__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only subtract a VinylTable from another VinylTable, not a {type(other)}"
            )
        return other.__sub__(self)

    @ensure_output
    def __mul__(self, other) -> VinylTable:
        if not isinstance(other, VinylTable):
            raise ValueError(
                f"Can only multiply a VinylTable by another VinylTable, not a {type(other)}"
            )

        ## NOTE: for now, we are matching based on the same column names. Will add more advanced joins based on pk/fk later
        common_columns = list(set(self.tbl.columns) & set(other.tbl.columns))
        if len(common_columns) == 0:
            return self.tbl.cross_join(other.tbl)
        else:
            return self.tbl.outer_join(other, predicates=common_columns)

    @ensure_output
    def select(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
    ) -> VinylTable:
        # get adjusted col names
        vinyl_by = VinylColumnList(self.tbl, by, unique=True)

        # need a copy of sort columns with asc() and desc() applied for sort and one without for selection
        vinyl_sort = VinylSortColumnList(self.tbl, sort, unique=True, reverse=False)

        vinyl_by_and_sort = VinylColumnList(
            self.tbl, vinyl_by.queryable + vinyl_sort.unsorted, unique=True
        )

        vinyl_cols = VinylColumnList(self.tbl, cols, unique=True)

        if (
            len(vinyl_by.queryable) == 0
            and len(vinyl_sort.unsorted) == 0
            and fill is None
        ):
            out = self.tbl.select(*vinyl_cols.queryable)
            return out

        window_ = ibis.window(
            group_by=None if by is None else vinyl_by.queryable,
            order_by=None if sort is None else vinyl_sort.sorted,
            range=(window_bounds if window_type == WindowType.range else None),
            rows=(window_bounds if window_type == WindowType.rows else None),
        )

        source_cols_to_select = VinylColumnList(
            self.tbl, vinyl_by_and_sort.queryable + vinyl_cols.sources_as_deferred
        )

        # Adjust for window
        windowed = vinyl_cols.windowize(window_)
        adj_cols_to_select = vinyl_by_and_sort + windowed

        # make simple selection if no fill or by
        adj_window_ = ibis.window(
            group_by=None if by is None else vinyl_by.names_as_deferred,
            order_by=None if sort is None else vinyl_sort.names_as_deferred_sorted,
            range=(window_bounds if window_type == WindowType.range else None),
            rows=(window_bounds if window_type == WindowType.rows else None),
        )

        if fill is None or len(vinyl_by_and_sort.cols) == 0:
            return self.tbl.select(adj_cols_to_select.queryable)

        # build spine
        spine, rename_dict = _build_spine(self.tbl, vinyl_by_and_sort.queryable)
        fill_list = _adjust_fill_list(len(vinyl_cols.cols), fill)

        # union spine and preaggregated table
        unioned = _join_spine(self.tbl.select(source_cols_to_select.queryable), spine)
        unioned_cols = VinylColumnList(unioned, vinyl_cols.lambdaized, unique=True)
        unioned_cols.windowize(adj_window_, adj_object=True)

        # perform base window calculations
        new_adj_cols_to_select = VinylColumnList(
            unioned,
            vinyl_by.names_as_deferred
            + vinyl_sort.names_as_deferred
            + unioned_cols.queryable,
            unique=True,
        )  # recreate windows with original cols

        filled = unioned.select(*new_adj_cols_to_select.queryable).rename(rename_dict)

        if all([i == FillOptions.null for i in fill_list]):
            if vinyl_sort.cols == []:
                return filled
            return filled.order_by(vinyl_sort.names_as_deferred_sorted)

        for i, col in enumerate(vinyl_cols):
            # switch to VinylTable to access deferred method
            filled = VinylTable(
                filled._arg, self._conn_replace, self._twin_conn_replace
            )
            filled.mutable = self.mutable

            filled = filled._interpolate(
                col.name_as_deferred_resolved(filled),
                sort=vinyl_sort.names_as_deferred_resolved_sorted(filled),
                by=vinyl_by.names_as_deferred_resolved(filled),
                fill=fill_list[i],
            )
        final_sort_cols = vinyl_sort.names_as_deferred_resolved_sorted(
            filled
        ) + vinyl_by.names_as_deferred_resolved(filled)
        if final_sort_cols == []:
            return filled
        return filled.sort(final_sort_cols)

    @ensure_output
    def select_all(
        self,
        col_selector: column_type_all,
        f: Callable[[Any], Any] | list[Callable[[Any], Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        adj_cols = _process_multiple_select(self.tbl, col_selector, f, rename=rename)
        return self.select(adj_cols, by, sort, window_type, window_bounds, fill)

    @ensure_output
    def mutate(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
    ) -> VinylTable:
        vinyl_cols = VinylColumnList(self.tbl, cols)
        vinyl_by = VinylColumnList(self.tbl, by)
        vinyl_sort = VinylSortColumnList(self.tbl, sort, reverse=False)

        out = self.select_all(
            col_selector=[s.all()] + vinyl_cols.queryable,
            f=[lambda x: x] + [None] * len(vinyl_cols.queryable),
            by=vinyl_by.queryable,
            sort=vinyl_sort.sorted,
            window_type=window_type,
            window_bounds=window_bounds,
            rename=True,
        )
        return out

    @ensure_output
    def mutate_all(
        self,
        col_selector: column_type_all,
        f: Callable[..., Any] | list[Callable[..., Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        window_type: WindowType = WindowType.rows,
        window_bounds: tuple[int | None, int | None] = (None, None),
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        col_selector_list = (
            [col_selector] if not isinstance(col_selector, list) else col_selector
        )
        f_list = [f] if not isinstance(f, list) else f
        if len(f_list) == 1 and len(col_selector_list) > 1:
            f_list = f_list * len(col_selector_list)

        return self.select_all(
            col_selector=[s.all()]
            + col_selector_list,  # ibis handles renames automtically
            f=[lambda x: x, *f_list],
            by=by,
            sort=sort,
            window_type=window_type,
            window_bounds=window_bounds,
            rename=rename,
        )

    @ensure_output
    def aggregate(
        self,
        cols: column_type,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        fill: fill_type = None,
    ) -> VinylTable:
        vinyl_cols = VinylColumnList(self.tbl, cols)
        vinyl_by = VinylColumnList(self.tbl, by)
        vinyl_sort = VinylSortColumnList(self.tbl, sort, reverse=False)
        spine_cols = VinylColumnList(
            self.tbl, vinyl_by.queryable + vinyl_sort.unsorted, unique=True
        )
        unadj_tbl = self.tbl.aggregate(
            metrics=vinyl_cols.queryable, by=spine_cols.queryable, having=()
        )

        if fill is None or len(spine_cols.cols) == 0:
            # Don't support having argument. use filter instead
            if vinyl_sort.cols == []:
                return unadj_tbl
            else:
                return unadj_tbl.order_by(vinyl_sort.names_as_deferred_sorted)

        spine, rename_dict = _build_spine(self.tbl, spine_cols.queryable)
        if spine is None:
            return unadj_tbl

        adj_aggregate = unadj_tbl.rename(rename_dict)
        null_fill = _join_spine(
            adj_aggregate, spine
        )  # rejoin spine to ensure all by times/dates are present

        filled = null_fill
        fill_list = _adjust_fill_list(len(vinyl_cols.cols), fill)
        if all([i == FillOptions.null for i in fill_list]):
            if vinyl_sort.cols == []:
                return filled
            return filled.order_by(vinyl_sort.names_as_deferred_sorted)
        # handle factor

        for i, col in enumerate(vinyl_cols):
            # switch to VinylTable to access deferred method
            filled = VinylTable(
                filled._arg, self._conn_replace, self._twin_conn_replace
            )
            filled.mutable = self.mutable
            filled = filled._interpolate(
                col.name_as_deferred_resolved(filled),
                sort=vinyl_sort.names_as_deferred_resolved_sorted(filled),
                by=vinyl_by.names_as_deferred_resolved(filled),
                fill=fill_list[i],
            )
        final_sort_cols = vinyl_sort.names_as_deferred_resolved_sorted(
            filled
        ) + vinyl_by.names_as_deferred_resolved(filled)
        if final_sort_cols == []:
            return filled
        return filled.sort(final_sort_cols)

    def execution_helper(
        self: VinylTable, twin=False, limit: int | None = 10000
    ) -> ir.Table:
        unactivated_conn_replace = (
            self._twin_conn_replace if twin else self._conn_replace
        )
        if unactivated_conn_replace == {} and twin is False:
            raise ValueError(
                "No connection found. Please ensure your connection is correctly set up."
            )
        if unactivated_conn_replace == {} and twin is True:
            raise ValueError(
                "No twin connection found. Please ensure your twin connection is correctly set up."
            )

        # activate conn replace by calling the function
        active_conn_replace = {k: v() for k, v in unactivated_conn_replace.items()}

        replaced = self.tbl.op().replace(active_conn_replace).to_expr()
        replaced = replaced.limit(limit) if limit is not None else replaced

        # handle queries with multiple backends
        if len(replaced._find_backends()[0]) > 1:
            return unify_backends(replaced)
        return replaced

    def execute(
        self,
        format: Literal["pandas", "polars", "pyarrow", "torch"] = "pandas",
        twin: bool = False,
        limit: int | None = 10000,
    ) -> Any:
        replaced = self.execution_helper(twin, limit)
        if format == "pandas":
            return (
                replaced.to_pyarrow().to_pandas()
            )  # normal ibis execute produces type errors which this resolves
        elif format == "pyarrow":
            return replaced.to_pyarrow()
        elif format == "polars":
            try:
                import polars as pl
            except ImportError:
                raise ImportError(
                    "Polars not installed. Please install polars to use it."
                )
            return pl.from_arrow(replaced.to_pyarrow())
        elif format == "torch":
            return replaced.to_torch()
        else:
            raise ValueError(f"Output type {format} not supported")

    def save(
        self,
        path: str | Path,
        format: Literal["csv", "delta", "json", "parquet"] = "csv",
        limit=10000,
    ):
        replaced = self.execution_helper(limit)

        if format == "csv":
            replaced.to_csv(path)
        elif format == "delta":
            replaced.to_delta(path)
        elif format == "json":
            with open(path, "w") as f:
                f.write(replaced.execute().to_json(orient="records"))
        elif format == "parquet":
            replaced.to_parquet(path)
        else:
            raise ValueError(f"Output type {format} not supported")

    def pickle(self, path: str | Path, tbl_only=True):
        if tbl_only:
            pkl.dump(self.tbl, open(path, "wb"))
        else:
            pkl.dump(self, open(path, "wb"))
        return

    @ensure_output
    def aggregate_all(
        self,
        col_selector: column_type_all,
        f: Callable[[Any], Any] | list[Callable[[Any], Any] | None] | None,
        by: column_type | None = None,
        sort: column_type | None = None,  # add support for selectors later
        fill: fill_type = None,
        rename: bool = False,
    ) -> VinylTable:
        adj_cols = _process_multiple_select(self.tbl, col_selector, f, rename=rename)
        return self.aggregate(cols=adj_cols, by=by, sort=sort, fill=fill)

    @ensure_output
    def rename(self, rename_dict: dict[str, str]) -> VinylTable:
        return self.tbl.rename(rename_dict)

    @ensure_output
    def relocate(
        self,
        cols: column_type_without_dict,
        before: base_column_type | s.Selector | None = None,
        after: base_column_type | s.Selector | None = None,
    ) -> VinylTable:
        vinyl_cols = VinylColumnList(self.tbl, cols)
        vinyl_before = VinylColumnList(self.tbl, before) if before is not None else None
        vinyl_after = VinylColumnList(self.tbl, after) if after is not None else None
        before_names = vinyl_before.names if vinyl_before is not None else []
        after_names = vinyl_after.names if vinyl_after is not None else []

        helper = []
        for ls in [before_names, after_names]:
            if len(ls) > 1:
                helper.append(~s.c(*ls))
            elif len(ls) == 1:
                helper.append(ls[0])
            else:
                helper.append(None)

        return self.tbl.relocate(*vinyl_cols.names, before=helper[0], after=helper[1])

    @ensure_output
    def _interpolate(
        self,  # should already have had null fill applied
        cols: column_type | None = None,
        sort: column_type | None = None,
        by: column_type | None = None,
        fill: fill_type = FillOptions.null,  # can specify one of the fill options or a lambda function
    ) -> VinylTable:
        vinyl_cols = VinylColumnList(self.tbl, cols)
        vinyl_by = VinylColumnList(self.tbl, by)
        vinyl_sort = VinylSortColumnList(
            self.tbl, sort, reverse=True if fill == FillOptions.previous else False
        )
        out: VinylTable = self
        adj_fill = _adjust_fill_list(len(vinyl_cols.cols), fill)
        for i, vinyl_col in enumerate(vinyl_cols.cols):
            fill = adj_fill[i]
            if fill is None:
                continue
            elif fill == FillOptions.null:
                out = out.sort(vinyl_sort.sorted)
                # no need to fill further, should already have n
                continue
            elif callable(fill):
                # initial mutate to allow fill to operate on unaltered columns
                fill_col_name = f"fill_{vinyl_col.name}"
                filled_col = fill(vinyl_col.name_as_deferred_resolved(self))
                if not isinstance(filled_col, ir.Expr):
                    filled_col = ibis.literal(filled_col)

                out = out.mutate(
                    {fill_col_name: filled_col},
                    by=vinyl_by.names_as_deferred_resolved(out),
                    sort=vinyl_sort.names_as_deferred_resolved_sorted(out),
                )

                fill_col = VinylColumn(out.tbl, fill_col_name)
                out = out.mutate_all(
                    vinyl_col.name_as_deferred_resolved(out),
                    f=lambda x: x.fillna(fill_col.name_as_deferred_resolved(out)),
                    by=vinyl_by.names_as_deferred_resolved(out),
                    sort=vinyl_sort.sorted,
                )
                out = out.drop(fill_col.name_as_deferred_resolved(out))
            else:
                if sort is None:
                    raise ValueError(
                        "Must provide a sort column when using next or previous fill"
                    )
                helper_col_string = f"helper_{vinyl_col.name}"
                out = self.mutate(
                    {
                        helper_col_string: vinyl_col.name_as_deferred_resolved(
                            out
                        ).count()
                    },
                    by=vinyl_by.names_as_deferred_resolved(out),
                    sort=vinyl_sort.names_as_deferred_resolved_sorted(out),
                    fill=None,
                    window_type=WindowType.rows,
                    window_bounds=(None, 0),
                )
                helper_col = VinylColumn(out.tbl, helper_col_string)
                out = out.mutate(
                    {vinyl_col.name: vinyl_col.name_as_deferred_resolved(out).max()},
                    by=vinyl_by.names_as_deferred_resolved(out)
                    + [helper_col.name_as_deferred_resolved(out)],
                )
                out = out.drop(helper_col.name_as_deferred_resolved(out))
                out = out.sort(vinyl_sort.names_as_deferred_resolved_sorted(out))

        return out

    def chart(
        self,
        geoms: geom
        | list[
            Any
        ],  # always will be a geom, but listing any here to avoid import of lets-plot unless necessary
        x: ir.Value | None,
        y: ir.Value | None = None,
        color: ir.Value | None = None,
        fill: ir.Value | None = None,
        size: ir.Value | None = None,
        alpha: ir.Value | None = None,
        facet: ir.Value | list[ir.Value] | None = None,
        coord_flip: bool = False,
    ):
        from vinyl.lib.chart import BaseChart

        return BaseChart(
            geoms, self, x, y, color, fill, size, alpha, facet, coord_flip
        ).show()

    ## modified version of ibis.sql that modifies sqlglot behavior to (1) minimize parsing and (2) optimize if requested
    def to_sql(self, dialect="duckdb", optimized=False, formatted=True) -> str:
        expr = self.tbl
        if dialect is None:
            try:
                backend = expr._find_backend()
            except com.IbisError:
                # default to duckdb for sqlalchemy compilation because it supports
                # the widest array of ibis features for SQL backends
                backend = ibis.duckdb
            else:
                sqlglot_dialect = getattr(backend, "_sqlglot_dialect", backend.name)
        else:
            try:
                backend = getattr(ibis, dialect)
            except AttributeError:
                raise ValueError(f"Unknown dialect {dialect}")
            else:
                sqlglot_dialect = getattr(backend, "_sqlglot_dialect", dialect)

        sql = backend._to_sql(expr)
        if optimized:
            return optimize(sql, dialect=sqlglot_dialect, rules=RULES).sql(
                dialect=sqlglot_dialect, pretty=True
            )
        elif formatted:
            parsed = sqlglot.parse(sql, dialect=sqlglot_dialect)[0]
            if parsed is not None:
                return parsed.sql(dialect=sqlglot_dialect, pretty=True)
            else:
                raise ValueError("Could not parse SQL")

        return sql

    @ensure_output
    def distinct(
        self,
        on: column_type_without_dict | None = None,
        keep: Literal["first", "last"] | None = "first",
    ) -> VinylTable:
        if on is not None:
            on = VinylColumnList(self.tbl, on, unique=True).names
        return self.tbl.distinct(on=on, keep=keep)

    @ensure_output
    def drop(self, cols: column_type_without_dict) -> VinylTable:
        vinyl_cols = VinylColumnList(self.tbl, cols)
        return self.select(~s._to_selector(vinyl_cols.names))

    @ensure_output
    def dropna(
        self,
        on: column_type_without_dict | None = None,
        how: Literal["any", "all"] = "any",
    ) -> VinylTable:
        if on is not None:
            on = VinylColumnList(self.tbl, on, unique=True).names
        return self.tbl.dropna(subset=on, how=how)

    @ensure_output
    def sort(
        self,
        by: column_type | None = None,
    ) -> VinylTable:
        if by is None:
            return self.tbl
        return self.tbl.order_by(by)

    @ensure_output
    def limit(self, n: int | None, offset: int = 0) -> VinylTable:
        return self.tbl.limit(n, offset)

    @ensure_output
    def join(
        self,
        right: Any,  # will be a VinylTable, but pydantic can't validate
        predicates: (
            str
            | Sequence[
                str
                | ir.BooleanColumn
                | Literal[True]
                | Literal[False]
                | tuple[
                    str | ir.Column,
                    str | ir.Column,
                ]
            ]
        ) = (),
        how: Literal[
            "inner",
            "left",
            "outer",
            "right",
            "semi",
            "anti",
            "any_inner",
            "any_left",
            "left_semi",
        ] = "inner",
        lname: str = "",
        rname: str = "{name}_right",
    ) -> VinylTable:
        return ibis.join(
            left=self.tbl,
            right=right.tbl,
            predicates=predicates,
            how=how,
            lname=lname,
            rname=rname,
        )

    @ensure_output
    def union(self, *rest: Any, distinct: bool = False) -> VinylTable:
        if len(rest) == 0:
            return self

        return reduce(lambda x, y: _union_two(x.tbl, y.tbl, distinct), rest, self)

    @ensure_output
    def filter(self, conditions: boolean_column_type) -> VinylTable:
        return self.tbl.filter(conditions)

    def filter_all(
        self,
        col_selector: column_type_all,
        condition_f: Callable[..., Any] | list[Callable[..., Any] | None] | None,
        condition_type: Literal["and", "or"] = "and",
    ) -> VinylTable:
        adj_cols = _process_multiple_select(
            self.tbl, col_selector, condition_f, rename=False
        )
        if condition_type == "and":
            return self.filter(adj_cols)
        else:
            for i, col in enumerate(adj_cols):
                if i == 0:
                    combined_cond = col
                else:
                    combined_cond |= col
            return self.filter(combined_cond)

    @ensure_output
    def count(
        self, where: base_boolean_column_type | None = None, distinct: bool = False
    ) -> VinylTable:
        vinyl_where = VinylColumn(self.tbl, where)
        if distinct:
            return self.tbl.nunique(where=vinyl_where.col).as_table()
        return self.tbl.count(where=vinyl_where.col).as_table()

    def schema(self) -> VinylSchema:
        ibis_schema = VinylSchema(self.tbl.schema())
        return ibis_schema

    def get_name(self):
        return self.tbl.get_name()

    @property
    def columns(self):
        return self.tbl.columns

    @classmethod
    def create_from_schema(cls, schema: Schema, name: str | None = None, conn=None):
        return cls(ibis.table(schema, name=name)._arg)


# Add pydantic validations. Need to do after function definition to make sure return class (VinylTable) is available to pydantic
VinylTable.select = validate(VinylTable.select)  # type: ignore[method-assign]
VinylTable.select_all = validate(VinylTable.select_all)  # type: ignore[method-assign]
VinylTable.mutate = validate(VinylTable.mutate)  # type: ignore[method-assign]
VinylTable.mutate_all = validate(VinylTable.mutate_all)  # type: ignore[method-assign]
VinylTable.aggregate = validate(VinylTable.aggregate)  # type: ignore[method-assign]
VinylTable.aggregate_all = validate(VinylTable.aggregate_all)  # type: ignore[method-assign]
VinylTable.rename = validate(VinylTable.rename)  # type: ignore[method-assign]
VinylTable.relocate = validate(VinylTable.relocate)  # type: ignore[method-assign]
VinylTable._interpolate = validate(VinylTable._interpolate)  # type: ignore[method-assign]
VinylTable.distinct = validate(VinylTable.distinct)  # type: ignore[method-assign]
VinylTable.drop = validate(VinylTable.drop)  # type: ignore[method-assign]
VinylTable.dropna = validate(VinylTable.dropna)  # type: ignore[method-assign]
VinylTable.sort = validate(VinylTable.sort)  # type: ignore[method-assign]
VinylTable.limit = validate(VinylTable.limit)  # type: ignore[method-assign]
VinylTable.join = validate(VinylTable.join)  # type: ignore[method-assign]
VinylTable.union = validate(VinylTable.union)  # type: ignore[method-assign]
VinylTable.filter = validate(VinylTable.filter)  # type: ignore[method-assign]
VinylTable.filter_all = validate(VinylTable.filter_all)  # type: ignore[method-assign]
VinylTable.count = validate(VinylTable.count)  # type: ignore[method-assign]



from collections.abc import Hashable
from typing import Any

import ibis.expr.operations as ops
import networkx as nx
from grandalf.graphs import Edge, Graph, Vertex  # type: ignore
from grandalf.layouts import SugiyamaLayout  # type: ignore
from netext.edge_rendering.arrow_tips import ArrowTip
from netext.edge_rendering.modes import EdgeSegmentDrawingMode
from netext.edge_routing.modes import EdgeRoutingMode
from netext.geometry.point import FloatPoint  # type: ignore
from netext.layout_engines.engine import G, LayoutEngine
from netext.node_rasterizer import JustContent
from netext.textual.widget import GraphView
from rich import print
from rich.table import Table
from rich.text import Text
from textual.app import ComposeResult
from textual.widgets import Footer

from vinyl.lib.utils.graphics import TurntableTextualApp, adjust_type


class GrandalfView:
    w = 0
    h = 0
    xy = (0, 0)


def _create_vertex_horizontal(node: Hashable, data: dict[str, Any]) -> Vertex:
    v = Vertex(node)

    # The API is a bit weird that it assumes to just add some members externally.
    v.view = GrandalfView()
    v.view.h = data["_netext_node_buffer"].layout_width * 5
    v.view.w = data["_netext_node_buffer"].layout_height * 2
    return v


class GrandalfLRSugiyamaLayout(LayoutEngine[G]):
    """Layout engine that uses the grandalf library to layout the graph using the Sugiyama algorithm.

    Multiple components will be placed next to each other.
    """

    def __call__(self, g: G) -> dict[Hashable, FloatPoint]:
        vertices = {
            node: _create_vertex_horizontal(node, data)
            for node, data in g.nodes(data=True)
        }
        edges = [Edge(vertices[u], vertices[v]) for u, v in g.edges]
        graph = Graph(vertices.values(), edges)

        # TODO Open up settings to netext
        for c in graph.components():
            sug = SugiyamaLayout(c)
            sug.init_all()
            sug.draw(3)
        # Rescale back, but leave a bit more space to avoid overlaps in the
        # terminal coordinate space.
        result = dict()
        y_offset = 0
        for c in graph.components():
            component = {
                v.data: FloatPoint(y=v.view.xy[0] / 4 + y_offset, x=v.view.xy[1] / 6)
                for v in c.sV
            }
            y_offset = max([v.view.xy[0] / 4 + y_offset + v.view.w for v in c.sV])
            result.update(component)
        return result


def _render(n, d, s):
    t = Table(title=str(n))
    node = d["node"]
    print(n)

    t.add_column("Name")
    t.add_column("Type")
    t.add_column("Key")

    schema = node.schema()
    types = schema.types

    for i, (name, type) in enumerate(schema.items()):
        field = getattr(node._source_class, name)
        if field.primary_key:
            key = "PK"
            to_bold = True
        elif field.foreign_key is not None:
            source = field.foreign_key.op().find(ops.UnboundTable)[0].name
            col_name = field.foreign_key.get_name()
            key = Text("FK", style="blue") + Text(
                f" {source}.{col_name}", style="white"
            )
            name = Text(name, style="blue")
            to_bold = False
        else:
            key = ""
            to_bold = False

        adj_type = adjust_type(types[i])

        if to_bold:
            t.add_row(name, adj_type, key, style="on bright_black")
        else:
            t.add_row(name, adj_type, key)

    return t


def create_erd_app(g: nx.DiGraph) -> GraphView:
    nx.set_node_attributes(g, _render, "$content-renderer")
    nx.set_node_attributes(g, JustContent(), "$shape")
    nx.set_edge_attributes(g, EdgeRoutingMode.ORTHOGONAL, "$edge-routing-mode")
    nx.set_edge_attributes(
        g, EdgeSegmentDrawingMode.BOX_ROUNDED, "$edge-segment-drawing-mode"
    )
    nx.set_edge_attributes(g, ArrowTip.ARROW, "$end-arrow-tip")

    class GraphviewApp(TurntableTextualApp):
        def compose(self) -> ComposeResult:
            yield GraphView(g, zoom=1.2, layout_engine=GrandalfLRSugiyamaLayout())
            yield Footer()

    return GraphviewApp()



import ast
import os
import re


class ClassInfo:
    def __init__(self, name, excluded_attributes=[]):
        self.name = name
        self.excluded_attrs = excluded_attributes
        self.attributes_dict = {}

    def add_attribute(self, key, value):
        if key not in self.excluded_attrs:
            self.attributes_dict[key] = value


class ClassFinder(ast.NodeVisitor):
    def __init__(self, excluded_attributes=[]):
        self.classes = []
        self.excluded_attrs = excluded_attributes

    def visit_ClassDef(self, node):
        class_info = ClassInfo(node.name, self.excluded_attrs)

        # Look for assignments that occur directly in the class body
        for body_item in node.body:
            # Only handle type-annotated assignments with a value
            if isinstance(body_item, ast.AnnAssign) and body_item.value is not None:
                if isinstance(body_item.target, ast.Name):
                    # Add the attribute along with its value's AST
                    class_info.add_attribute(
                        body_item.target.id, ast.unparse(body_item.value)
                    )

        self.classes.append(class_info)
        self.generic_visit(node)  # Continue visiting child nodes


def find_classes_and_attributes(file_path):
    if not os.path.exists(file_path):
        return {}

    with open(file_path, "r") as file:
        source = file.read()

    tree = ast.parse(source)
    finder = ClassFinder(
        [
            "_table",
            "_unique_name",
            "_path",
            "_twin_path",
            "_row_count",
            "_schema",
            "_database",
        ]
    )
    finder.visit(tree)

    return finder.classes[0].attributes_dict


# more accurate
def get_imports_from_file_ast(file_path):
    with open(file_path, "r") as file:
        source = file.read()

    imports = ""
    tree = ast.parse(source)
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            imports += ast.unparse(node)
        elif isinstance(node, ast.ImportFrom):
            imports += ast.unparse(node)
        elif isinstance(node, ast.ClassDef):
            break
        imports += "\n"

    return imports


# more literal
def get_imports_from_file_regex(file_path, regex="@source"):
    with open(file_path, "r") as file:
        source = file.read()

    parts = re.split(regex, source, re.MULTILINE)

    if len(parts) == 1:
        return ""

    return parts[0]



import os

import duckdb
import ibis
import typer

from vinyl.infra.pg_proxy.server import create as create_pg_proxy
from vinyl.lib.definitions import load_project_defs
from vinyl.lib.metric import MetricStore
from vinyl.lib.project import Project

project_cli = typer.Typer(pretty_exceptions_show_locals=False)


@project_cli.command("deploy")
def deploy():
    """Deploy a Vinyl project"""
    defs = load_project_defs()
    if not os.path.exists(".vinyl"):
        os.makedirs(".vinyl")
    if os.path.exists(".vinyl/vinyl.duckdb"):
        os.remove(".vinyl/vinyl.duckdb")

    conn = ibis.duckdb.connect(".vinyl/vinyl.duckdb")
    project = Project(resources=defs.resources, models=defs.models)
    for model in project.models:
        table = model()
        if isinstance(table, MetricStore):
            # we dont support this yet
            continue

        conn.create_table(model.__name__, table.execute().to_pyarrow())


@project_cli.command("serve")
def serve():
    """Serve a Vinyl project"""
    bv_host = "0.0.0.0"
    bv_port = 5433
    if not os.path.exists(".vinyl/vinyl.duckdb"):
        raise RuntimeError("No database found in .vinyl/vinyl.duckdb")

    conn = duckdb.connect(".vinyl/vinyl.duckdb")
    print("(Turntable) Using DuckDB database at .vinyl/vinyl.duckdb")
    print("(Turntable) Tables:")
    tables = conn.execute("show tables;").fetchall()
    for table in tables:
        print(f"- {table[0]}")
    server = create_pg_proxy(conn, (bv_host, bv_port))
    ip, port = server.server_address
    print(f"\nListening on {ip}:{port}")
    try:
        server.serve_forever()
    finally:
        server.shutdown()
        conn.close()



import copy
import warnings
from collections.abc import Iterable
from typing import Any, Callable, Sequence, Set

import ibis
import ibis.expr.datatypes as dt
import ibis.expr.types as ir
import ibis.selectors as s
from ibis import _
from ibis.common.deferred import Deferred
from pydantic import BaseModel, ConfigDict

from vinyl.lib.column import (
    VinylColumn,
    VinylColumnList,
    _process_cols_arg,
    column_type,
)
from vinyl.lib.constants import DIMENSION_LABEL, METRIC_LABEL, SCHEMA_LABEL, TS_COL_NAME
from vinyl.lib.enums import FillOptions, WindowType
from vinyl.lib.table import VinylTable
from vinyl.lib.table_methods import _adjust_fill_list, _join_with_removal
from vinyl.lib.temporal import set_timezone
from vinyl.lib.utils.text import split_interval_string


class Metric(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,  # prevents error tied to ibis types not being fully pydantic compatible
        validate_assignment=(
            False  # don't actually validate yet, will ship in future release
        ),
    )

    tbl: VinylTable
    name: str = ""  # allows pydantic to still work
    ts: Callable[..., Any]
    agg: Callable[..., Any]
    by: column_type
    fill: FillOptions | Callable[..., Any]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # set name
        self.name = VinylColumn(self.tbl.tbl, self.agg, passthrough_deferred=True).name

        # ensure ts is a timestamp
        self.ts = set_timezone(self.tbl.tbl, self.ts)

        # ensure ts name is set to TS_COL_NAME constant
        ts_orig = copy.deepcopy(self.ts)
        self.ts = lambda t: ts_orig(t).name(TS_COL_NAME)

    @property
    def dimension_names(self) -> Set[str]:
        if self.by is None:
            return set()
        vinyl_by = VinylColumnList(self.tbl.tbl, self.by, passthrough_deferred=True)
        return set(vinyl_by.names)

    def to_dict(self):
        return {
            "tbl": self.tbl,
            "name": self.name,
            "agg": self.agg,
            "ts": self.ts,
            "by": self.by,
            "fill": self.fill,
        }


def get_dimension_names(metrics: Sequence[Metric]) -> Set[str]:
    return set().union(*[m.dimension_names for m in metrics])


class MetricStore:
    all_metrics_dict: dict[str, Metric] = dict()
    all_dimension_names: set[str] = set()

    mutable: bool
    default_tbl: VinylTable | None
    trailing_intervals: list[ibis.interval] | None
    metrics_dict: dict[str, Metric]
    dimension_names: set[str]

    # key will always be a vinylTable, but can't use it here due to a circular reference
    tbl_dict: dict[VinylTable, dict[str, Any]] = {}

    def __init__(
        self,
        metrics: Sequence[Metric] = [],
        default_tbl: VinylTable | None = None,
    ):
        self.mutable = False
        self.default_tbl = default_tbl
        self.trailing_intervals = None
        self.metrics_dict = {m.name: m for m in metrics}
        self.dimension_names = get_dimension_names(metrics)

        self.tbl_dict = {}
        for m in metrics:
            if m.tbl not in self.tbl_dict:
                self.tbl_dict[m.tbl] = {}
                self.tbl_dict[m.tbl][METRIC_LABEL] = set()
                self.tbl_dict[m.tbl][DIMENSION_LABEL] = set()
                self.tbl_dict[m.tbl][SCHEMA_LABEL] = {}

            self.tbl_dict[m.tbl][METRIC_LABEL].add(m.name)
            self.tbl_dict[m.tbl][DIMENSION_LABEL].update(m.dimension_names)
            col_helper_it = VinylColumn(m.tbl.tbl, m.agg, passthrough_deferred=True)
            self.tbl_dict[m.tbl][SCHEMA_LABEL].setdefault(
                col_helper_it.name, col_helper_it.type
            )
            self.tbl_dict[m.tbl][SCHEMA_LABEL][TS_COL_NAME] = dt.Timestamp()
            if m.by is not None and isinstance(m.by, Iterable):
                for dim in VinylColumnList(m.tbl.tbl, m.by, passthrough_deferred=True):
                    self.tbl_dict[m.tbl][SCHEMA_LABEL].setdefault(dim.name, dim.type)

        MetricStore.address_naming_conflicts(self.metrics_dict)
        MetricStore.append(self.metrics_dict)

        self.set_columns()

    def copy(self):
        return MetricStore(list(self.metrics_dict.values()), self.default_tbl)

    def set_columns(self):
        # set columns
        for i, name in enumerate(["ts", *self.dimension_names, *self.metrics_dict]):
            col = getattr(_, name)
            setattr(self, name, col)
            setattr(self, f"_{i}", col)

    def clear_columns(self):
        for name in self.__dict__.copy():
            if name not in [
                "mutable",
                "default_tbl",
                "trailing_intervals",
                "metrics_dict",
                "dimension_names",
                "tbl_dict",
            ]:
                delattr(self, name)

    def reset_columns(self):
        self.clear_columns()
        self.set_columns()

    def __add__(self, other):
        if not isinstance(other, MetricStore):
            raise ValueError(
                f"Can only add two MetricStores together, not a {type(other)}"
            )
        if self.default_tbl is None:
            return other
        if other.default_tbl is None:
            return self

        if self.default_tbl != other.default_tbl:
            raise ValueError(
                "Can't add two MetricStores with different default tables. Use a join (e.g. tbl1 * tbl2) instead."
            )
        combined_metrics = list(self.metrics_dict.values()) + list(
            other.metrics_dict.values()
        )

        return MetricStore(
            combined_metrics,
            self.default_tbl,
        )

    def __radd__(self, other: "MetricStore") -> "MetricStore":
        return other.__add__(self)

    # Make callable so that @model wrapper works
    def __call__(self):
        return self

    # Make mutable when using in a context manager
    def __enter__(self):
        # Create a copy of the original object and make the object mutable
        new = MetricStore(list(self.metrics_dict.values()), self.default_tbl)
        new.mutable = True
        return new

    def __exit__(self, exc_type, exc_value, traceback):
        # Exit logic here
        pass

    def select(
        self,
        cols: Sequence[Deferred] | dict[str, Deferred],
        trailing: list[int | None] = [None],
    ) -> Any:
        if self.default_tbl is None:
            raise ValueError("Can't select from an empty MetricStore.")
        cols_adj = _process_cols_arg(
            self.default_tbl.tbl, cols, passthrough_deferred=True
        )
        return MetricSelect(self, cols=cols_adj, intervals=trailing).select()

    def metric(
        self,
        tbl: VinylTable,
        cols: ir.Scalar | Sequence[ir.Scalar] | dict[str, ir.Scalar],
        ts: ir.TimestampValue,
        by: Sequence[ir.Value] = [],
        fill: FillOptions = FillOptions.null,  # all metrics have fill
    ):
        vinyl_cols = VinylColumnList(tbl.tbl, cols, passthrough_deferred=True)
        if len(vinyl_cols.cols) == 0:
            raise ValueError("Must provide at least one metric to metric function")

        vinyl_by = VinylColumnList(tbl.tbl, by, passthrough_deferred=True)
        vinyl_ts = VinylColumn(tbl.tbl, ts, passthrough_deferred=True)

        fill_list = _adjust_fill_list(len(vinyl_cols.cols), fill)

        mets = []
        for i, v in enumerate(vinyl_cols):
            met = Metric(
                tbl=tbl,
                agg=v.lambdaized,
                ts=vinyl_ts.lambdaized,
                by=vinyl_by.lambdaized,
                fill=fill_list[i],
            )
            mets.append(met)
        out = MetricStore(mets, tbl)
        combined = self + out
        combined.reset_columns()
        if self.mutable:
            self.__dict__ = combined.__dict__
            self.__dict__["mutable"] = True  # reset mutability to True

        return combined

    def trailing(self, intervals: list[tuple[int, str]] | list[str]):
        if isinstance(intervals[0], str):
            intervals = [split_interval_string(i) for i in intervals]

        ## arguments here should be like (1, "d")
        self.trailing_invervals = [ibis.interval(*i) for i in intervals]

    @classmethod
    def append(cls, new_metrics_dict):
        cls.all_metrics_dict.update(new_metrics_dict)
        cls.all_dimension_names = cls.all_dimension_names | get_dimension_names(
            new_metrics_dict.values()
        )

    @classmethod
    def address_naming_conflicts(cls, new_metrics_dict):
        # set warnings for metric name conflicts
        all_intersections = set(new_metrics_dict.keys()) & set(
            cls.all_metrics_dict.keys()
        )
        nontrivial_intersections = set()
        for i in all_intersections:
            new_metrics_entry = new_metrics_dict[i]
            cls_metrics_entry = cls.all_metrics_dict[i]
            if hash(new_metrics_entry.agg(new_metrics_entry.tbl.tbl)) != hash(
                cls_metrics_entry.agg(cls_metrics_entry.tbl.tbl)
            ):
                nontrivial_intersections.add(i)

        if any(nontrivial_intersections):
            warning_txt = f"\n\nMetric(s) {nontrivial_intersections} already exist in the global metric store. The new metric(s) will overwrite the old metric(s).\n\n Here's how the formulas will change:\n"
            for k, v in new_metrics_dict.items():
                if k in nontrivial_intersections:
                    warning_txt += f"- {k}: {cls.all_metrics_dict[k].agg(cls.all_metrics_dict[k].tbl.tbl)} -> {v.agg(v.tbl.tbl)}\n"
            warnings.warn(message=warning_txt)

        # raise error when metric name conflicts with dimension name
        new_all_metric_names = set(new_metrics_dict.keys()) | set(
            cls.all_metrics_dict.keys()
        )
        new_all_dimension_names = get_dimension_names(new_metrics_dict.values()) | set(
            cls.all_dimension_names
        )

        if any(new_all_metric_names & new_all_dimension_names):
            raise ValueError(
                f"\n\nMetric name(s) {new_all_metric_names} conflict with dimension name(s) {new_all_dimension_names}. Please rename the metric(s) or dimension(s)."
            )


class MetricSelect:
    def __init__(
        self,
        MetricStore: MetricStore,
        cols: list[Deferred],
        intervals: list[int | None] = [None],
    ):
        self.metric_store = MetricStore
        self.intervals = intervals
        self.ts: list[Deferred] = []
        self.dimensions: list[Deferred] = []
        self.temp_dimensions: list[Deferred] = []
        self.temp_metrics: list[Deferred] = []

        # helper to select metrics from the same tbl together as long as they have the same ts
        self.tbl_ts_metrics_to_select: dict[VinylTable, list[Deferred]] = {}

        # helper to store raw metric tables
        self.raw_metric_tbls: list[VinylTable] = []

        self.all_exprs = cols

        # create combined schema
        for i, vals in enumerate(list(self.metric_store.tbl_dict.values())):
            if i == 0:
                self.combined_schema = VinylTable.create_from_schema(vals[SCHEMA_LABEL])
            else:
                self.combined_schema += VinylTable.create_from_schema(
                    vals[SCHEMA_LABEL]
                )

    def _process_unaltered_col(self, col, col_name):
        if col_name in self.metric_store.dimension_names:
            self.dimensions.append(col)
        elif col_name in self.metric_store.metrics_dict:
            met = self.metric_store.metrics_dict[col_name]
            key_it = met.tbl.mutate(met.ts(met.tbl))  # use ts column as unique
            self.tbl_ts_metrics_to_select.setdefault(key_it, set())
            self.tbl_ts_metrics_to_select[key_it].add(col_name)
        elif col_name == TS_COL_NAME:
            raise ValueError(
                "Can't select the raw ts column. Consider using the .truncate() or .bucket() method"
            )
        else:
            raise ValueError(f"{col_name} is not a valid column name")

    def _process_derived_col(self, col, col_name):
        vinyl_col = VinylColumn(
            self.combined_schema.tbl, col, passthrough_deferred=True
        )
        sources = vinyl_col.sources
        source_types = []
        for source in sources:
            if source in self.metric_store.dimension_names:
                source_types.append(DIMENSION_LABEL)
            elif source in self.metric_store.metrics_dict:
                source_types.append(METRIC_LABEL)
            elif source == TS_COL_NAME:
                source_types.append(TS_COL_NAME)
            else:
                raise ValueError(f"{source} is not a valid column name")

            if len(set(source_types)) > 1:
                raise ValueError(
                    "Can't mix metrics and dimensions in the same expression"
                )
            elif source_types[0] == TS_COL_NAME:
                self.ts.append(col)
            elif source_types[0] == DIMENSION_LABEL:
                self.temp_dimensions.append(col)
                # ensure that sources are added to dimensions so derivation works later
                self.dimensions.append(source)
            elif source_types[0] == METRIC_LABEL:
                self.temp_metrics.append(col)
                # ensure that sources are added to metrics so derivation works later
                met_source = self.metric_store.metrics_dict[source]
                key_it = met_source.tbl.mutate(
                    met_source.ts(met_source.tbl)
                )  # use ts column as unique so that the same tables with different ts columns are aggregated separately
                self.tbl_ts_metrics_to_select.setdefault(key_it, set())
                self.tbl_ts_metrics_to_select[key_it].add(source)
            else:
                raise ValueError(f"{col_name} is not a valid column name")

    def process_col(self, col):
        vinyl_col = VinylColumn(
            self.combined_schema.tbl, col, passthrough_deferred=True
        )
        col_name = vinyl_col.name
        mets_to_select = []
        for i in self.tbl_ts_metrics_to_select.values():
            mets_to_select.extend(list(i))
        cur_col_names_helper = [
            d
            for d in list(self.dimensions)
            + list(self.temp_dimensions)
            + list(self.temp_metrics)
            + mets_to_select
        ]
        cur_col_names = VinylColumnList(
            self.combined_schema.tbl, cur_col_names_helper, passthrough_deferred=True
        ).names

        # skip if column is already in dimensions or metrics already selected (including derived)
        if col_name in cur_col_names:
            return

        if vinyl_col.is_unaltered:
            self._process_unaltered_col(vinyl_col.lambdaized, col_name)

        else:
            self._process_derived_col(vinyl_col.lambdaized, col_name)

    def _build_final_tbls(self):
        for interval in self.intervals:
            raw_metric_tbl_interval = []
            mets_rename_dict = {}
            # prepare raw metric tables
            for tbl, met_names in self.tbl_ts_metrics_to_select.items():
                tbl = tbl.copy()  # copy tbl to ensure no mutability issues
                mets_it = [
                    self.metric_store.metrics_dict[name].agg for name in met_names
                ]
                fill_it = [
                    self.metric_store.metrics_dict[name].fill for name in met_names
                ]
                vinyl_cols = VinylColumnList(
                    tbl.tbl, mets_it, passthrough_deferred=True
                )
                source_cols = vinyl_cols.sources_as_deferred_resolved(tbl.tbl)
                vinyl_dims = VinylColumnList(
                    tbl.tbl, self.dimensions, passthrough_deferred=True
                )
                source_dims = vinyl_dims.names_as_deferred
                vinyl_ts = VinylColumn(tbl.tbl, self.ts, passthrough_deferred=True)
                source_ts = [vinyl_ts.name]
                ts_original_source = vinyl_ts.sources

                if interval is None:
                    tbl = tbl.select(
                        source_cols,
                        by=[col(tbl.tbl) for col in self.dimensions],
                        sort=[col(tbl.tbl) for col in self.ts],
                    )

                    out_it = tbl.aggregate(
                        cols=[met(tbl.tbl) for met in mets_it],
                        by=source_dims,
                        sort=source_ts,
                        fill=fill_it,
                    )
                else:
                    tbl = tbl.aggregate_all(
                        col_selector=source_cols,
                        f=lambda x: x.collect(),
                        by=[col(tbl.tbl) for col in self.dimensions],
                        sort=[col(tbl.tbl) for col in self.ts],
                        fill=FillOptions.null,  # trailing metrics only support null fill,
                    )
                    collected_tbl_with_trailing = tbl.select_all(
                        col_selector=vinyl_cols.sources_as_deferred_resolved(tbl.tbl),
                        f=lambda x: x.collect().flatten(),
                        by=[dim(tbl) for dim in source_dims],
                        sort=[vinyl_ts.name_as_deferred_resolved(tbl)],
                        window_type=WindowType.rows,
                        window_bounds=(-interval, 0),
                        rename=False,
                    )

                    ## NOTE: to replace once Vinyl mutate function is written
                    unnested = collected_tbl_with_trailing.mutate_all(
                        col_selector=[s.of_type(dt.Array)],
                        f=[lambda y: y.unnest()],
                        rename=False,
                    ).rename(
                        {v: source_ts[i] for i, v in enumerate(ts_original_source)}
                    )  # rename ts to make sure transformation is available for the fill in next step

                    out_it = unnested.aggregate(
                        cols=[met(unnested.tbl) for met in mets_it],
                        by=[dim(unnested.tbl) for dim in source_dims],
                        sort=[col(unnested.tbl) for col in self.ts],
                        fill=FillOptions.null,
                    )

                    mets_rename_dict.update()

                raw_metric_tbl_interval.append(out_it)

            # join together
            joined_raw_metric_tbl_interval = raw_metric_tbl_interval[0]
            for tbl in raw_metric_tbl_interval[1:]:
                # Takes connection replace of first table, this will fail if there are two different connections in the list
                joined_raw_metric_tbl_interval = VinylTable(
                    _join_with_removal(
                        joined_raw_metric_tbl_interval.tbl, tbl.tbl
                    )._arg,
                    _conn_replace=joined_raw_metric_tbl_interval._conn_replace,
                    _twin_conn_replace=joined_raw_metric_tbl_interval._twin_conn_replace,
                )

            # calculate derived dimensions and metrics
            for col in self.temp_dimensions + self.temp_metrics:
                joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.mutate(
                    col(joined_raw_metric_tbl_interval.tbl)
                )

            # reorder to match original requested order
            cols_to_select = [
                col.resolve(joined_raw_metric_tbl_interval)
                for col in VinylColumnList(
                    self.combined_schema.tbl, self.all_exprs, passthrough_deferred=True
                ).names_as_deferred
            ]
            joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.select(
                cols_to_select
            )

            # rename trailing metrics
            if interval is not None:
                to_rename = [
                    i.get_name()
                    for i in s.where(
                        lambda x: x.get_name() not in source_ts + vinyl_dims.names
                    ).expand(joined_raw_metric_tbl_interval.tbl)
                ]
                joined_raw_metric_tbl_interval = joined_raw_metric_tbl_interval.rename(
                    {f"{v}_{interval}": v for v in to_rename}
                )
            self.raw_metric_tbls.append(joined_raw_metric_tbl_interval)

    def select(self):
        for col in self.all_exprs:
            self.process_col(col)

        # build raw metric tables
        self._build_final_tbls()

        # join raw metric tables
        joined_raw_metric_tbl = self.raw_metric_tbls[0]
        for tbl in self.raw_metric_tbls[1:]:
            # Takes connection replace of first table, this will fail if there are two different connections in the list
            joined_raw_metric_tbl = VinylTable(
                _join_with_removal(joined_raw_metric_tbl.tbl, tbl.tbl)._arg,
                _conn_replace=joined_raw_metric_tbl._conn_replace,
                _twin_conn_replace=joined_raw_metric_tbl._twin_conn_replace,
            )

        # sort outputs
        final = joined_raw_metric_tbl.sort(s.all())

        return final



import os

import typer
from rich.console import Console
from rich.table import Table
from tqdm import tqdm

from vinyl.lib.connect import (
    DatabaseFileConnector,
    Source,
    _DatabaseConnector,
    _FileConnector,
    _TableConnector,
)
from vinyl.lib.definitions import load_project_defs
from vinyl.lib.project import Project, load_project_module
from vinyl.lib.source import _get_twin_relative_path
from vinyl.lib.utils.ast import find_classes_and_attributes, get_imports_from_file_regex
from vinyl.lib.utils.files import create_dirs_with_init_py
from vinyl.lib.utils.functions import with_modified_env
from vinyl.lib.utils.text import make_python_identifier

console = Console()


sources_cli = typer.Typer(pretty_exceptions_show_locals=False)


@sources_cli.command("list")
def list_sources(tables: bool = False):
    """Caches sources to a local directory (default: .turntable/sources)"""
    defs = load_project_defs()
    project = Project(resources=defs.resources, models=defs.models)

    table = Table("Name", "Resource", "Location", title="Sources")
    for source in project.get_source_objects():
        table.add_row(
            f"[bold]{source.name}[bold]",
            f"[grey70]{source.parent_resource.def_.__name__}[grey70]",
            f"[grey70]{source.location}[grey70]",
        )
    console.print(table)


def table_to_python_class(table_name):
    return "".join([word.capitalize() for word in table_name.split("_")])


def source_to_class_string(
    source: Source,
    saved_attributes: dict[str, str],
    generate_twin: bool = False,
    root_path: str = None,
    sample_size: int = 1000,
):
    if not root_path:
        os.path.dirname(load_project_module().__file__)

    class_name = table_to_python_class(source.name)
    class_body = f'    _table = "{source.name}"\n'
    pr = source.parent_resource

    if pr is None:
        raise ValueError("Source must have a parent resource.")

    if isinstance(pr.connector, _TableConnector):
        tbl = pr.connector.get_table(source.location)
        class_body += f'    _unique_name = "{pr.name}.{class_name}"\n'

    elif isinstance(pr.connector, _DatabaseConnector):
        # source is a database
        database, schema = source.location.split(".")
        class_body += f'    _unique_name = "{pr.name}.{source.location}.{class_name}"\n'
        class_body += f'    _schema = "{schema}"\n'
        class_body += f'    _database = "{database}"\n'

        database, schema = source.location.split(".")
        tbl = pr.connector.get_table(database, schema, source.name)
        class_body += f'    _twin_path = "{_get_twin_relative_path(pr.name, database, schema)}"\n\n'
        # need row count if using local files (since sampling is done live from the file)

    else:
        raise NotImplementedError(
            f"Connector type {type(pr.connector)} is not yet supported"
        )

    # need row count if using local files (since sampling is done live from the file)
    if isinstance(pr.connector, _FileConnector):
        if isinstance(pr.connector, DatabaseFileConnector):
            # in this case, location is not the real path, but the database and schema, but we can use the path from the connector
            path = os.path.relpath(pr.connector.path, root_path)
        else:
            path = os.path.relpath(source.location, root_path)
        class_body += f'    _path = "{path}"\n'
        if not pr.connector.remote:
            # in this case, we will not be caching the table, so we need the row_count
            class_body += f"    _row_count = {tbl.count().execute()}\n\n"

    if source.schema is None:
        raise ValueError(f"Schema for {source.name} is not available")

    for col_name, col_type in source.schema.items():
        base = f"    {col_name.lower().replace(' ', '_')}: t.{col_type.__repr__()}"
        if col_name in saved_attributes:
            base += f" = {saved_attributes[col_name]}"
        class_body += f"{base}\n"

    out = f"""class {class_name}:
{class_body}
"""
    return out


def get_save_dir(sources_path: str, source: Source) -> str:
    if source.parent_resource is None:
        raise ValueError("Source must have a parent resource.")
    if isinstance(source.parent_resource.connector, _TableConnector):
        # source is a file
        return os.path.join(sources_path, source.parent_resource.name)
    if isinstance(source.parent_resource.connector, _DatabaseConnector):
        # source is a database
        identifers = [
            make_python_identifier(str_) for str_ in source.location.split(".")
        ]
        return os.path.join(sources_path, source.parent_resource.name, *identifers)
    raise NotImplementedError(
        f"Connector type {type(source.parent_resource.connector)} is not yet supported"
    )


@sources_cli.command("generate")
def generate_sources(
    twin: bool = typer.Option(
        False, "--generate_twin", "-t", help="exported name of the model"
    ),
    resources: list[str] = typer.Option(
        [], "--resource", "-r", help="resource names to select"
    ),
):
    """Generates schema files for sources"""

    @with_modified_env("NO_MODELS_VINYL", "True")
    def run_fn():
        defs = load_project_defs()
        if len(resources) == 0:
            project = Project(resources=defs.resources)
        else:
            project = Project(
                resources=[r for r in defs.resources if r.__name__ in resources]
            )
        root_path = os.path.dirname(load_project_module().__file__)
        sources = project.get_source_objects(with_schema=True)

        sources_path = os.path.join(root_path, "sources")
        create_dirs_with_init_py(sources_path)

        for source in sources:
            save_dir = get_save_dir(sources_path, source)
            create_dirs_with_init_py(save_dir)
            file_path = os.path.join(save_dir, f"{source.name}.py")
            saved_attributes = find_classes_and_attributes(file_path)
            saved_imports = (
                get_imports_from_file_regex(file_path)
                if saved_attributes != {}
                else None
            )
            with open(os.path.join(save_dir, f"{source.name}.py"), "w+") as f:
                if saved_imports:
                    f.write(saved_imports)
                else:
                    f.write("# type: ignore\n")  # prevents pylance errors tied to Ibis
                    f.write("from vinyl import source\n")
                    f.write("from vinyl import types as t\n\n")
                    f.write(
                        f"from {source.parent_resource.def_.__module__} import {source.parent_resource.def_.__name__}\n\n\n"
                    )
                f.write(f"@source(resource={source.parent_resource.name})\n")
                f.write(
                    source_to_class_string(
                        source, saved_attributes, root_path=root_path
                    )
                )

        if twin:
            msg = "generating twins... " if len(sources) > 1 else "generating twin... "
            for source in tqdm(
                sources,
                msg,
                unit="source",
            ):
                pr = source.parent_resource
                if isinstance(pr.connector, _DatabaseConnector):
                    database, schema = source.location.split(".")
                    pr.connector.generate_twin(
                        os.path.join(root_path, _get_twin_relative_path(pr.name)),
                        database,
                        schema,
                        source.name,
                    )
                elif isinstance(pr.connector, _TableConnector):
                    # doesn't actually generate a file, just returns the path
                    pr.connector.generate_twin(source.location)

        print(f"Generated {len(sources)} sources at {sources_path}")

    run_fn()



from __future__ import annotations

import builtins
from typing import TypeAlias

import ibis.expr.datatypes as types  # noqa: F401
import ibis.selectors as sel  # noqa: F401
import rich
from ibis import _  # noqa: F401

from vinyl.lib.asset import (  # noqa: F401
    metric,
    model,
    resource,  # noqa: F401
)
from vinyl.lib.definitions import load_defs  # noqa: F401
from vinyl.lib.enums import FillOptions  # noqa: F401,
from vinyl.lib.expressions import case, if_else  # noqa: F401
from vinyl.lib.field import Field  # noqa: F401
from vinyl.lib.metric import MetricStore  # noqa: F401
from vinyl.lib.operators import ilike, is_, isin, isnt, like, notin  # noqa: F401
from vinyl.lib.source import source  # noqa: F401
from vinyl.lib.table import VinylTable  # noqa: F401

builtins.print = rich.print  # type: ignore

T: TypeAlias = VinylTable
M: TypeAlias = MetricStore



import ibis
import ibis.expr.types as ir


def case(
    pairs: list[tuple[ir.BooleanValue, ir.Value]], default: ir.Value | None = None
) -> ir.Value:
    if len(pairs) == 0:
        raise ValueError("At least one pair must be provided")
    elif len(pairs) == 1:
        return ibis.ifelse(pairs[0][0], pairs[0][1], default)

    out = ibis.case()
    for pair in pairs:
        out = out.when(pair[0], pair[1])
    out = out.else_(default)
    return out.end()


def if_else(
    condition: ir.BooleanValue, true_value: ir.Value, false_value: ir.Value
) -> ir.Value:
    return ibis.ifelse(condition, true_value, false_value)



import functools
import inspect
from functools import wraps
from typing import Any, Callable

from vinyl.lib.metric import MetricStore
from vinyl.lib.table import VinylTable
from vinyl.lib.utils.functions import validate


@validate
def base(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    publish: bool = False,
    tags: str | list[str] | None = None,
    asset_type: str = "model",
):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args):
            if not isinstance(deps, list):
                deps_adj = [deps]
            else:
                deps_adj = deps
            # Map the positional arguments to the new names
            aliases = [
                param.name for param in inspect.signature(func).parameters.values()
            ]
            if len(aliases) != len(deps_adj):
                raise Exception("Wrong number of arguments")

            new_kwargs = {}
            for i, alias in enumerate(aliases):
                # TODO: this is a hack to avoid mutable sources and models from being impacted downstream
                # we should eventually clean this up as it's not a sustainable solution
                dep_it = deps_adj[i]()

                if isinstance(dep_it, (VinylTable, MetricStore)):
                    par = dep_it.copy()

                else:
                    raise ValueError(
                        f"Dependencies must be VinylTable or MetricStore, not {type(dep_it)}"
                    )

                par.mutable = True
                new_kwargs[alias] = par

            # Call the original function with the new arguments
            return func(**new_kwargs)

        setattr(wrapper, f"_is_vinyl_{asset_type}", True)

        return wrapper

    return decorator


@validate
def model(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    publish: bool = False,
    tags: str | list[str] | None = None,
):
    return base(deps, publish, tags, asset_type="model")


@validate
def metric(
    deps: object | Callable[..., Any] | list[object | Callable[..., Any]],
    publish: bool = False,
    tags: str | list[str] | None = None,
):
    return base(deps, publish, tags, asset_type="metric_store")


def resource(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        return f(*args, **kwargs)

    decorated_function._is_vinyl_resource = True
    return decorated_function



from ibis import Schema

from vinyl.lib.utils.graphics import print_schema


class VinylSchema:
    schema: Schema

    def __init__(self, schema: Schema):
        self.schema = schema

    def __str__(self):
        return self.schema.__str__().replace("ibis.Schema {", "vinyl.Schema {")

    def __repr__(self):
        return self.schema.__repr__().replace("ibis.Schema {", "vinyl.Schema {")

    def items(self):
        return self.schema.items()

    @property
    def names(self):
        return self.schema.names

    @property
    def types(self):
        return self.schema.types

    def __rich__(self):
        return print_schema(self.schema)



import datetime
import hashlib
import io
import json
import logging
import os
import random
import socketserver
import struct
from typing import Dict, Optional

from vinyl.infra.pg_proxy.core import (
    BVType,
    Connection,
    Extension,
    QueryResult,
    Session,
)
from vinyl.infra.pg_proxy.rewrite import Rewriter

logger = logging.getLogger(__name__)

NULL_BYTE = b"\x00"


class ServerResponse:
    """Byte codes for server responses in the PG wire protocol."""

    AUTHENTICATION_REQUEST = b"R"
    BACKEND_KEY_DATA = b"K"
    BIND_COMPLETE = b"2"
    CLOSE_COMPLETE = b"3"
    COMMAND_COMPLETE = b"C"
    DATA_ROW = b"D"
    EMPTY_QUERY_RESPONSE = b"I"
    ERROR_RESPONSE = b"E"
    NO_DATA = b"n"
    NOTICE_RESPONSE = b"N"
    PARAMETER_DESCRIPTION = b"t"
    PARAMETER_STATUS = b"S"
    PARSE_COMPLETE = b"1"
    PORTAL_SUSPENDED = b"s"
    READY_FOR_QUERY = b"Z"
    ROW_DESCRIPTION = b"T"


class TransactionStatus:
    IDLE = b"I"
    IN_TRANSACTION = b"T"
    IN_FAILED_TRANSACTION = b"E"


# Client commands
class ClientCommand:
    """Byte codes for client commands in the PG wire protocol."""

    BIND = b"B"
    CLOSE = b"C"
    DESCRIBE = b"D"
    EXECUTE = b"E"
    FLUSH = b"H"
    QUERY = b"Q"
    PARSE = b"P"
    PASSWORD_MESSAGE = b"p"
    SYNC = b"S"
    TERMINATE = b"X"


def _time_to_microseconds(t):
    # Convert hours, minutes, seconds, and microseconds to microseconds
    hours_to_microseconds = t.hour * 60 * 60 * 1e6
    minutes_to_microseconds = t.minute * 60 * 1e6
    seconds_to_microseconds = t.second * 1e6
    microseconds = t.microsecond
    total_microseconds = (
        hours_to_microseconds
        + minutes_to_microseconds
        + seconds_to_microseconds
        + microseconds
    )
    return int(total_microseconds)


def _micros_since_2000(dt):
    micros = (
        dt - datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
    ).total_seconds() * 1000000
    return int(micros)


PG_UNKNOWN = (705, str)
BVTYPE_TO_PGTYPE = {
    BVType.NULL: (-1, lambda v: None),
    BVType.ARRAY: (2277, lambda v: "{" + ",".join(v) + "}", None),
    BVType.BIGINT: (20, str, lambda r: int.to_bytes(r, 8, "big")),
    BVType.BOOL: (
        16,
        lambda v: "true" if v else "false",
        lambda r: b"\x01" if r else b"\x00",
    ),
    BVType.BYTES: (17, lambda v: "\\x" + v.hex(), lambda r: r),
    BVType.DATE: (
        1082,
        lambda v: v.isoformat(),
        lambda r: int.to_bytes((r.toordinal() - 730120), 4, "big"),
    ),
    BVType.DECIMAL: (1700, str),
    BVType.FLOAT: (701, str, lambda r: struct.pack("!d", r)),
    BVType.INTEGER: (23, str, lambda r: int.to_bytes(r, 4, "big")),
    BVType.INTEGERARRAY: (1007, lambda v: "{" + ",".join(v) + "}"),
    BVType.INTERVAL: (
        1186,
        lambda v: f"{v.days} days {v.seconds} seconds {v.microseconds} microseconds",
    ),
    BVType.JSON: (114, lambda v: json.dumps(v)),
    BVType.STRINGARRAY: (1009, lambda v: "{" + ",".join(v) + "}"),
    BVType.TEXT: (25, str, lambda r: r.encode("utf-8")),
    BVType.TIME: (
        1083,
        lambda v: v.isoformat(),
        lambda r: int.to_bytes(_time_to_microseconds(r), 8, "big"),
    ),
    BVType.TIMESTAMP: (
        1114,
        lambda v: v.isoformat().replace("T", " "),
        lambda r: int.to_bytes(_micros_since_2000(r), 8, "big"),
    ),
}


class BVBuffer(object):
    """A helper for reading and writing bytes in the format the PG wire protocol expects."""

    def __init__(self, stream=None):
        if not stream:
            stream = io.BytesIO()
        self.stream = stream

    def read_bytes(self, n) -> bytes:
        return self.stream.read(n)

    def read_byte(self) -> bytes:
        return self.read_bytes(1)

    def read_int16(self) -> int:
        data = self.read_bytes(2)
        return struct.unpack("!h", data)[0]

    def read_uint32(self) -> int:
        data = self.read_bytes(4)
        return struct.unpack("!I", data)[0]

    def read_int32(self) -> int:
        data = self.read_bytes(4)
        return struct.unpack("!i", data)[0]

    def write_bytes(self, value: bytes):
        self.stream.write(value)

    def write_byte(self, value):
        self.stream.write(struct.pack("!c", value))

    def write_int16(self, value: int):
        self.stream.write(struct.pack("!h", value))

    def write_int32(self, value: int):
        self.stream.write(struct.pack("!i", value))

    def write_string(self, value):
        self.stream.write(value.encode() if isinstance(value, str) else value)
        self.stream.write(b"\x00")

    def get_value(self) -> bytes:
        return self.stream.getvalue()


class BVContext:
    """Manages the state of a single connection to the server."""

    def __init__(
        self, session: Session, rewriter: Optional[Rewriter], params: Dict[str, str]
    ):
        self.session = session
        self.rewriter = rewriter
        self.params = params
        self.process_id = random.randint(0, 2**32 - 1)
        self.secret_key = random.randint(0, 2**32 - 1)
        self.stmts = {}
        self.portals = {}
        self.result_cache = {}
        self.has_error = False
        self.authenticated = False
        self.salt = None

    def get_hashed_password(self, auth: dict) -> str:
        user = self.params["user"]
        password = auth[user]
        first = hashlib.md5(password.encode("utf-8") + user.encode("utf-8")).hexdigest()
        return "md5" + hashlib.md5(first.encode("utf-8") + self.salt).hexdigest()

    def mark_error(self):
        self.has_error = True

    def transaction_status(self):
        if self.session.in_transaction():
            if self.has_error:
                return TransactionStatus.IN_FAILED_TRANSACTION
            else:
                return TransactionStatus.IN_TRANSACTION
        return TransactionStatus.IDLE

    def execute_sql(self, sql: str, params=None, result_fmt=None) -> QueryResult:
        logger.info("Input SQL: " + sql)
        if self.rewriter:
            sql = self.rewriter.rewrite(sql)
            logger.info("Rewritten SQL: " + sql)
        qr = self.session.execute_sql(sql, params)
        if qr.has_results():
            if result_fmt and len(result_fmt) != qr.column_count():
                qr.result_format = [result_fmt[0]] * qr.column_count()
            else:
                qr.result_format = result_fmt
        return qr

    def describe_portal(self, name: str) -> QueryResult:
        stmt, params, result_fmt = self.portals[name]
        sql, param_oids = self.stmts[stmt]
        # todo: parse params? LIMIT 0?
        query_result = self.execute_sql(sql=sql, params=params, result_fmt=result_fmt)
        self.result_cache[name] = query_result
        return query_result

    def describe_statement(self, name: str) -> QueryResult:
        sql, param_oids = self.stmts[name]
        # TODO: create default params from param_oids
        return self.execute_sql(sql)

    def execute_portal(self, name: str) -> QueryResult:
        if name in self.result_cache:
            query_result = self.result_cache[name]
            del self.result_cache[name]
            return query_result
        else:
            stmt, params, result_fmt = self.portals[name]
            sql, param_oids = self.stmts[stmt]
            # parse the params?
            qr = self.execute_sql(sql=sql, params=params, result_fmt=result_fmt)
            return qr

    def add_statement(self, name: str, sql: str, param_oids: list[int]):
        self.stmts[name] = (sql, param_oids)

    def close_statement(self, name: str):
        del self.stmts[name]

    def add_portal(
        self, name: str, stmt: str, params: Dict[str, str], result_formats: list[int]
    ):
        self.portals[name] = (stmt, params, result_formats)

    def close_portal(self, name: str):
        del self.portals[name]

    def flush(self):
        pass

    def sync(self):
        if self.has_error:
            self.has_error = False


class ProxyServerHandler(socketserver.StreamRequestHandler):
    def handle(self):
        self.r = BVBuffer(self.rfile)
        ctx = None
        try:
            ctx = self.handle_startup(self.server.conn)
            if ctx:
                self.server.ctxts[ctx.process_id] = ctx
            while ctx:
                type_code = self.r.read_byte()
                if not type_code or type_code == ClientCommand.TERMINATE:
                    # we're done
                    break

                msglen = self.r.read_uint32()
                if msglen > 4:
                    payload = self.r.read_bytes(msglen - 4)
                else:
                    payload = None

                if not ctx.authenticated:
                    if type_code == ClientCommand.PASSWORD_MESSAGE:
                        self.handle_md5_password(ctx, payload)
                    else:
                        raise Exception("Not authenticated")
                elif type_code == ClientCommand.QUERY:
                    self.handle_query(ctx, payload)
                elif type_code == ClientCommand.PARSE:
                    self.handle_parse(ctx, payload)
                elif type_code == ClientCommand.BIND:
                    self.handle_bind(ctx, payload)
                elif type_code == ClientCommand.DESCRIBE:
                    self.handle_describe(ctx, payload)
                elif type_code == ClientCommand.EXECUTE:
                    self.handle_execute(ctx, payload)
                elif type_code == ClientCommand.CLOSE:
                    self.handle_close(ctx, payload)
                elif type_code == ClientCommand.SYNC:
                    ctx.sync()
                    self.send_ready_for_query(ctx)
                elif type_code == ClientCommand.FLUSH:
                    ctx.flush()
                else:
                    raise Exception("Unknown type_code: %s" % type_code)
        except Exception as e:
            logger.exception(e)
            self.send_error(e)

        if ctx:
            self.server.conn.close_session(ctx.session)
            del self.server.ctxts[ctx.process_id]
            ctx = None

    def handle_startup(self, conn: Connection) -> BVContext:
        msglen = self.r.read_uint32() - 4
        code = self.r.read_uint32()
        if code == 80877103:  ## SSL request
            self.send_notice()
            return self.handle_startup(conn)
        elif code == 196608:  # Protocol 3.0
            msg = [
                x.decode("utf-8")
                for x in self.r.read_bytes(msglen - 4).split(NULL_BYTE)
            ]
            params = dict(zip(msg[::2], msg[1::2]))
            logger.info("Client connection params: %s", params)
            ctx = BVContext(conn.create_session(), self.server.rewriter, params)
            self.send_auth_request(ctx)
            return ctx
        elif code == 80877102:  ## Cancel request
            process_id, secret_key = self.r.read_uint32(), self.r.read_uint32()
            ctx = self.server.ctxts.get(process_id)
            if ctx and ctx.secret_key == secret_key:
                self.server.conn.close_session(ctx.session)
                del self.server.ctxts[ctx.process_id]
                ctx = None
            return None
        else:
            raise Exception(f"Unsupported startup message: {code}")

    def send_auth_request(self, ctx: BVContext):
        if self.server.auth is None:
            self.send_authentication_ok()
            self.handle_post_auth(ctx)
        else:
            self.send_authentication_md5(ctx)

    def send_authentication_ok(self):
        self.wfile.write(
            struct.pack("!cii", ServerResponse.AUTHENTICATION_REQUEST, 8, 0)
        )

    def send_authentication_md5(self, ctx: BVContext):
        ctx.salt = os.urandom(4)
        self.wfile.write(
            struct.pack("!cii", ServerResponse.AUTHENTICATION_REQUEST, 12, 5)
        )
        self.wfile.write(ctx.salt)

    def handle_md5_password(self, ctx: BVContext, payload: bytes):
        client_side = payload.decode("utf-8").rstrip("\x00")
        server_side = ctx.get_hashed_password(self.server.auth)
        if client_side == server_side:
            self.send_authentication_ok()
            self.handle_post_auth(ctx)
        else:
            self.send_error("Invalid password")

    def handle_post_auth(self, ctx: BVContext):
        self.send_parameter_status(self.server.conn.parameters())
        self.send_backend_key_data(ctx)
        self.send_ready_for_query(ctx)
        ctx.authenticated = True
        return

    def handle_query(self, ctx: BVContext, payload: bytes):
        logger.debug("Handle query")
        decoded = payload.decode("utf-8").rstrip("\x00")
        try:
            # JSON payloads signal that we should use extensions
            if req := Extension.check_json(decoded):
                method = req.get("method")
                extension = self.server.extensions.get(method)
                if not extension:
                    raise Exception("Unknown method: " + str(method))
                else:
                    query_result = extension.apply(req.get("params"), ctx.session)
            else:
                query_result = ctx.execute_sql(decoded)
        except Exception as e:
            self.send_error(e)
            self.send_ready_for_query(ctx)
            return

        if not query_result:
            raise Exception("No query result for query: " + decoded)
        elif query_result.has_results():
            self.send_row_description(query_result)
            row_count = self.send_data_rows(query_result)
            self.send_command_complete("SELECT %d\x00" % row_count)
        else:
            status = query_result.status()
            self.send_command_complete(f"{status}\x00")
        self.send_ready_for_query(ctx)

    def handle_parse(self, ctx: BVContext, payload: bytes):
        logger.debug("Handling parse")
        ba = bytearray(payload)
        stmt_idx = ba.index(NULL_BYTE)
        stmt = ba[:stmt_idx].decode("utf-8")
        query_idx = ba.index(NULL_BYTE, stmt_idx + 1)
        sql = ba[stmt_idx + 1 : query_idx].decode("utf-8")
        logger.debug("Parsed statement: %s", sql)
        buf = BVBuffer(io.BytesIO(ba[query_idx + 1 :]))
        num_params = buf.read_int16()
        param_oids = []
        for i in range(num_params):
            param_oids.append(buf.read_int32())
        ctx.add_statement(stmt, sql, param_oids)
        self.send_parse_complete()

    def handle_bind(self, ctx: BVContext, payload: bytes):
        logger.debug("Handling bind")
        ba = bytearray(payload)
        portal_idx = ba.index(NULL_BYTE)
        portal = ba[:portal_idx].decode("utf-8")
        stmt_idx = ba.index(NULL_BYTE, portal_idx + 1)
        stmt = ba[portal_idx + 1 : stmt_idx].decode("utf-8")
        buf = BVBuffer(io.BytesIO(ba[stmt_idx + 1 :]))
        # First param format stuff...
        num_formats = buf.read_int16()
        formats = []
        for i in range(num_formats):
            formats.append(buf.read_int16())
        # ... then the actual param values
        num_params = buf.read_int16()
        if num_formats < num_params:
            if formats:
                formats = [formats[0]] * num_params
            else:
                formats = [0] * num_params
        params = []
        for i in range(num_params):
            nb = buf.read_int32()
            v = buf.read_bytes(nb)
            if formats[i] == 0:
                decoded = v.decode("utf-8")
                if decoded.startswith("{") and decoded.endswith("}"):
                    params.append(decoded[1:-1].split(","))
                else:
                    params.append(decoded)
            else:
                # TODO: I shouldn't be always assuming these are always
                # ints but I can live with it for now
                params.append(int.from_bytes(v, "big"))
        logger.debug("Bind params: %s", params)
        # now expected result formats
        num_result_formats = buf.read_int16()
        result_formats = []
        for i in range(num_result_formats):
            result_formats.append(buf.read_int16())
        ctx.add_portal(portal, stmt, params, result_formats)
        self.send_bind_complete()

    def handle_describe(self, ctx: BVContext, payload: bytes):
        logger.debug("Handling describe")
        ba = bytearray(payload)
        describe_type = ba[0]
        query_result = None
        if describe_type == ord("P"):
            portal = ba[1 : len(ba) - 1].decode("utf-8")
            try:
                query_result = ctx.describe_portal(portal)
            except Exception as e:
                self.send_error(e, ctx)
                return
        elif describe_type == ord("S"):
            stmt = ba[1 : len(ba) - 1].decode("utf-8")
            try:
                query_result = ctx.describe_statement(stmt)
            except Exception as e:
                self.send_error(e, ctx)
                return
        else:
            raise Exception(f"Unknown describe type: {describe_type}")
        if query_result.has_results():
            self.send_row_description(query_result)

    def handle_execute(self, ctx: BVContext, payload: bytes):
        logger.debug("Handling execute")
        if ctx.has_error:
            logger.info("Skipping execute due to previous error")
            return
        ba = bytearray(payload)
        portal_idx = ba.index(NULL_BYTE)
        portal = ba[:portal_idx].decode("utf-8")
        limit = struct.unpack("!i", ba[portal_idx + 1 : portal_idx + 5])[0]
        query_result = None
        try:
            query_result = ctx.execute_portal(portal)
        except Exception as e:
            self.send_error(e, ctx)
            return
        if query_result.has_results():
            row_count = self.send_data_rows(query_result, limit)
            self.send_command_complete("SELECT %d\x00" % row_count)
        else:
            status = query_result.status()
            self.send_command_complete(f"{status}\x00")

    def handle_close(self, ctx: BVContext, payload: bytes):
        logger.debug("Handling close")
        close_type = payload[0]
        if close_type == ord("S"):
            ctx.close_statement(payload[1:-1].decode("utf-8"))
        elif close_type == ord("P"):
            ctx.close_portal(payload[1:-1].decode("utf-8"))
        else:
            raise Exception(f"Unknown close type: {close_type}")
        self.send_close_complete()
        self.send_ready_for_query(ctx)

    def send_row_description(self, query_result: QueryResult):
        buf = BVBuffer()
        for i in range(query_result.column_count()):
            name, bvtype = query_result.column(i)
            oid = BVTYPE_TO_PGTYPE.get(bvtype, PG_UNKNOWN)[0]
            buf.write_string(name)
            fmt = query_result.result_format[i] if query_result.result_format else 0
            buf.write_bytes(struct.pack("!ihihih", 0, 0, oid, 0, -1, fmt))
        out = buf.get_value()
        sig = struct.pack(
            "!cih",
            ServerResponse.ROW_DESCRIPTION,
            len(out) + 6,
            query_result.column_count(),
        )
        self.wfile.write(sig + out)

    def send_data_rows(self, query_result: QueryResult, limit: int = 0) -> int:
        cnt = 0
        converters = []
        for i in range(query_result.column_count()):
            bvtype = query_result.column(i)[1]
            pgtype = BVTYPE_TO_PGTYPE.get(bvtype, PG_UNKNOWN)
            if not query_result.result_format or query_result.result_format[i] == 0:
                converters.append((pgtype[1], True))
            else:
                converters.append((pgtype[2], False))
        for row in query_result.rows():
            buf = BVBuffer()
            for j in range(query_result.column_count()):
                r = row[j]
                if r is None:
                    buf.write_int32(-1)
                else:
                    converter, do_encode = converters[j]
                    v = converter(r)
                    if do_encode:
                        v = v.encode("utf-8")
                    buf.write_int32(len(v))
                    buf.write_bytes(v)
            out = buf.get_value()
            row_sig = struct.pack(
                "!cih",
                ServerResponse.DATA_ROW,
                len(out) + 6,
                query_result.column_count(),
            )
            self.wfile.write(row_sig + out)
            cnt += 1
            if limit > 0 and cnt >= limit:
                break
        return cnt

    def send_error(self, exception, ctx: Optional[BVContext] = None):
        estr = str(exception)
        logger.error(estr)
        buf = BVBuffer()
        buf.write_byte(b"M")
        buf.write_string(estr)
        buf.write_byte(NULL_BYTE)
        out = buf.get_value()
        err_sig = struct.pack("!ci", ServerResponse.ERROR_RESPONSE, len(out) + 4)
        self.wfile.write(err_sig + out)
        if ctx:
            ctx.mark_error()

    def send_notice(self):
        self.wfile.write(ServerResponse.NOTICE_RESPONSE)

    def send_backend_key_data(self, ctx):
        self.wfile.write(
            struct.pack(
                "!ciII",
                ServerResponse.BACKEND_KEY_DATA,
                12,
                ctx.process_id,
                ctx.secret_key,
            )
        )

    def send_ready_for_query(self, ctx: Optional[BVContext]):
        logger.debug("Sending ready for query")
        status = ctx.transaction_status() if ctx else TransactionStatus.IDLE
        self.wfile.write(struct.pack("!cic", ServerResponse.READY_FOR_QUERY, 5, status))

    def send_parameter_status(self, params: Dict[str, str]):
        for name, value in params.items():
            buf = BVBuffer()
            buf.write_string(name)
            buf.write_string(value)
            out = buf.get_value()
            psig = struct.pack("!ci", ServerResponse.PARAMETER_STATUS, len(out) + 4)
            self.wfile.write(psig + out)

    def send_parse_complete(self):
        self.wfile.write(struct.pack("!ci", ServerResponse.PARSE_COMPLETE, 4))

    def send_bind_complete(self):
        self.wfile.write(struct.pack("!ci", ServerResponse.BIND_COMPLETE, 4))

    def send_close_complete(self):
        self.wfile.write(struct.pack("!ci", ServerResponse.CLOSE_COMPLETE, 4))

    def send_command_complete(self, tag: str):
        buf = BVBuffer()
        buf.write_bytes(
            struct.pack("!ci", ServerResponse.COMMAND_COMPLETE, len(tag) + 4)
        )
        buf.write_bytes(tag.encode())
        out = buf.get_value()
        self.wfile.write(out)


class ProxyServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True

    def __init__(
        self,
        server_address,
        conn: Connection,
        *,
        rewriter: Optional[Rewriter] = None,
        extensions: list[Extension] = [],
        auth: Optional[Dict[str, str]] = None,
    ):
        super().__init__(server_address, ProxyServerHandler)
        self.conn = conn
        self.rewriter = rewriter
        self.extensions = {e.type(): e for e in extensions}
        self.ctxts = {}
        self.auth = auth

    def verify_request(self, request, client_address) -> bool:
        """Ensure all requests come from localhost until auth is in place"""
        return client_address[0] == "127.0.0.1" or "HOST" in os.environ



import dataclasses
import importlib
from typing import Any, Callable
import sys
import os
import ibis.expr.types as ir
import toml

from vinyl.lib.connect import _ResourceConnector
from vinyl.lib.utils.pkg import find_nearest_pyproject_toml_directory


def get_project_module_name():
    root_path = find_nearest_pyproject_toml_directory()

    with open(root_path, "r") as file:
        data = toml.load(file)

    return data["tool"]["vinyl"]["module_name"]


def load_project_module():
    # Backup the original sys.modules
    try:
        module_name = get_project_module_name()
    except KeyError:
        raise ValueError("Can't find a project pyproject.toml file.")

    sys.path.append(os.getcwd())
    imported_module = importlib.import_module(module_name)
    return imported_module


@dataclasses.dataclass
class Resource:
    name: str
    connector: _ResourceConnector
    def_: Any


class Project:
    resources: list[Resource]
    sources: list[ir.Table]
    models: list[Callable[..., Any]]
    metrics: list[Callable[..., Any]]

    def __init__(
        self,
        resources: list[Any],
        sources: list[ir.Table] | None = None,
        models: list[Any] | None = None,
        metrics: list[Any] | None = None,
    ):
        self.resources = [
            Resource(
                name=resource_def.__name__,
                def_=resource_def,
                connector=resource_def(),
            )
            for resource_def in resources
        ]
        if sources is not None:
            self.sources = sources
        if models is not None:
            self.models = models
        if metrics is not None:
            self.metrics = metrics

    def get_source_objects(self, with_schema=False) -> list[Any]:
        sources = []
        for resource in self.resources:
            try:
                resource_sources = resource.connector.list_sources(with_schema)
                for source in resource_sources:
                    source.parent_resource = resource
                    sources.append(source)
            except Exception as e:
                print(f"Error loading sources from {resource.name}: {e}")
                continue
        return sources

    def get_resource(self, resource_name: str):
        resources = [
            resource for resource in self.resources if resource.name == resource_name
        ]

        if len(resources) == 0:
            raise ValueError(f"Resource {resource_name} not found")

        resource = resources[0]

        return resource

    def get_source(self, source_id: str):
        sources = [source for source in self.sources if source.__name__ == source_id]

        if len(sources) == 0:
            raise ValueError(f"Source {source_id} not found")

        source = sources[0]

        return source

    def get_model(self, model_id: str):
        if self.models is None:
            raise ValueError("No models found")
        models = [model for model in self.models if model.__name__ == model_id]

        if len(models) == 0:
            raise ValueError(f"Model {model_id} not found")

        model = models[0]

        return model()

    def get_metric_store(self, metric_id: str):
        metrics = [metric for metric in self.metrics if metric.__name__ == metric_id]

        if len(metrics) == 0:
            raise ValueError(f"Metric {metric_id} not found")

        metric = metrics[0]

        return metric()

    def get_asset(self, asset_name: str):
        try:
            return self.get_resource(asset_name)
        except (AttributeError, ValueError):
            try:
                return self.get_source(asset_name)
            except (AttributeError, ValueError):
                try:
                    return self.get_model(asset_name)
                except (AttributeError, ValueError):
                    try:
                        self.get_metric_store(asset_name)
                    except (AttributeError, ValueError):
                        raise ValueError(f"Asset {asset_name} not found")



from datetime import datetime
from typing import Any, Callable

import ibis
import ibis.expr.types as ir
import pytz
from ibis.common.deferred import Deferred

from vinyl.lib.column import VinylColumn
from vinyl.lib.settings import PyProjectSettings


def get_utc_offset(timezone_str: str) -> float:
    # Create a timezone object
    tz = pytz.timezone(timezone_str)

    # Localize a datetime object to the specified timezone
    localized_dt = tz.localize(datetime.now())

    # Get the UTC offset in minutes
    utc_offset = localized_dt.utcoffset()
    if utc_offset is not None:
        return utc_offset.total_seconds() / 60

    raise ValueError(f"Invalid timezone: {timezone_str}")


## NOTE: This is a bit hacky. Some databases don't support timestamps with timezones effectively, so we use adjusted timestamps without timezones instead.
def set_timezone(
    tbl: Any,  # will be vinylTable but can't import here due to circular reference
    col: ir.Value | Callable[..., Any],
    to_tz: str | None = None,
    from_tz: str = "UTC",
) -> Deferred | ir.Value:
    # Check if col is a timestamp, otherwise raise an error
    vinyl_col = VinylColumn(tbl, col)

    if not vinyl_col.type.is_timestamp():
        raise TypeError("Column must be of type timestamp")

    # If no timezone is specified, check pyproject.toml
    to_tz = (
        PyProjectSettings().get_setting("tz")
        or str(datetime.now().astimezone().tzinfo)
        or "UTC"
    )

    # Create a timezone object
    utc_offset = get_utc_offset(to_tz)
    if from_tz != "UTC":
        utc_offset -= get_utc_offset(from_tz)

    # Cast to timestamp without tz (a bit hacky in ibis) and add the new timezone. Note this is a bit hacky in ibis. You can't cast to an instance of dt.Timestamp of it is deferred.
    col_as_dt = (
        col.cast(datetime) + ibis.interval(minutes=utc_offset)
        if isinstance(col, ir.Value) or isinstance(col, Deferred)
        else lambda x: col(x.tbl).cast(datetime) + ibis.interval(minutes=utc_offset)
    )
    return col_as_dt



from __future__ import annotations

import copy
from typing import Any, Callable, Sequence, TypeAlias

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis import selectors as s
from ibis.common.deferred import Deferred

from vinyl.lib.graph import VinylGraph
from vinyl.lib.utils.functions import validate

base_column_type: TypeAlias = str | ir.Value | Callable[..., Any]

base_boolean_column_type: TypeAlias = ir.BooleanValue | Callable[..., Any]
boolean_column_type: TypeAlias = (
    base_boolean_column_type | list[base_boolean_column_type]
)


column_type_without_dict: TypeAlias = (
    base_column_type | s.Selector | list[base_column_type | s.Selector]
)

column_type: TypeAlias = column_type_without_dict | dict[str, base_column_type]

column_type_all: TypeAlias = column_type_without_dict | list[column_type_without_dict]


@validate
def _append_name(col: ir.Value | Deferred, name: str | None = None) -> base_column_type:
    if name is None:
        return col
    try:  # check if v can be named, if it can't, wrap in coalesce
        col_named = col.name(name)
    except Exception:
        col_named = ibis.coalesce(col).name(name)

    return col_named


@validate
def _process_cols_arg(
    tbl: ir.Table,
    cols: column_type | None,
    names: list[str | None] | str | None = None,
    passthrough_deferred=False,
) -> list[ir.Value]:
    out: list[ir.Value] = []
    name = names[0] if isinstance(names, list) else names
    if cols is None:
        pass

    elif isinstance(cols, ir.Value):
        out = [_append_name(cols, name)]

    elif isinstance(cols, str):
        out = [_append_name(getattr(tbl, cols), name)]

    elif isinstance(cols, Deferred):
        if passthrough_deferred:
            out = [_append_name(cols, name)]
        else:
            out = [_append_name(cols.resolve(tbl), name)]

    elif callable(cols):
        if passthrough_deferred:
            out = [lambda t: _append_name(cols(t), name)]
        else:
            out = [_append_name(cols(tbl), name)]

    elif isinstance(cols, s.Selector):
        out = cols.expand(tbl)

    elif isinstance(cols, list):
        for col in cols:
            out.extend(
                _process_cols_arg(tbl, col, passthrough_deferred=passthrough_deferred)
            )

    elif isinstance(cols, dict):
        for name, col in cols.items():
            out.extend(
                _process_cols_arg(
                    tbl, col, name, passthrough_deferred=passthrough_deferred
                )
            )

    return out


class VinylColumn:
    def __init__(
        self,
        tbl: ir.Table,
        col: ir.Value | Sequence[ir.Value | s.Selector] | None,
        passthrough_deferred=False,
    ):
        self.tbl = tbl
        if col is None:
            self.col = None
        else:
            self.col = _process_cols_arg(
                self.tbl, col, passthrough_deferred=passthrough_deferred
            )[0]

    @property
    def name(self):
        if self.col is None:
            return None

        elif isinstance(self.col, str):
            return self.col

        elif isinstance(self.col, Deferred):
            return self.col.resolve(self.tbl).get_name()

        elif callable(self.col):
            return self.col(self.tbl).get_name()

        return self.col.get_name()

    @property
    def type(self):
        if self.col is None:
            return None

        elif isinstance(self.col, str):
            return getattr(self.tbl, self.col).resolve(self.tbl).type()

        elif isinstance(self.col, Deferred):
            return self.col.resolve(self.tbl).type()

        elif callable(self.col):
            return self.col(self.tbl).type()

        return self.col.type()

    @property
    def name_as_deferred(self):
        if self.col is None:
            return None
        return getattr(_, self.name)

    def name_as_deferred_resolved(self, tbl):
        if self.col is None:
            return None
        return getattr(tbl, self.name)

    @property
    def lambdaized(self):
        if isinstance(self.col, Deferred):
            self.col = self.col.resolve(self.tbl)
        elif callable(self.col):
            return self.col
        return lambda t: self.col.op().replace({self.tbl.op(): t.op()}).to_expr()

    @property
    def sources(self):
        if isinstance(self.col, Deferred):
            self.col = self.col.resolve(self.tbl)

        elif callable(self.col):
            self.col = self.col(self.tbl)

        return [i.name for i in self.col.op().find_topmost(ops.Field)]

    @property
    def nodes(self):
        graph = VinylGraph.new_init_from_expr(self.tbl.select(self.col))
        return graph.nodes()

    @property
    def is_unaltered(self):
        allowed = (ops.Field, ops.Relation)
        return all([isinstance(n, allowed) for n in self.nodes])

    @property
    def is_only_aliased(self):
        allowed = (ops.Field, ops.Relation, ops.Alias)
        return all([isinstance(n, allowed) for n in self.nodes])


class VinylColumnList:
    def __init__(
        self,
        tbl: ir.Table,
        cols: dict[str, Deferred | ir.Value]
        | list[Deferred | ir.Value | s.Selector]
        | Deferred
        | ir.Value
        | s.Selector
        | None,
        unique=False,
        passthrough_deferred=False,
    ):
        self.tbl = tbl
        self.unique = unique
        self.cols = []

        cols = _process_cols_arg(
            self.tbl, cols, passthrough_deferred=passthrough_deferred
        )
        if cols is None:
            pass
        elif isinstance(cols, list):
            for col in cols:
                if isinstance(col, s.Selector):
                    self.cols.extend(
                        [
                            VinylColumn(
                                tbl, c, passthrough_deferred=passthrough_deferred
                            )
                            for c in col.expand(self.tbl)
                        ]
                    )
                else:
                    self.cols.append(
                        VinylColumn(tbl, col, passthrough_deferred=passthrough_deferred)
                    )
        else:
            if isinstance(cols, s.Selector):
                self.cols = [
                    VinylColumn(tbl, c, passthrough_deferred=passthrough_deferred)
                    for c in cols.expand(self.tbl)
                ]
            else:
                self.cols = [
                    VinylColumn(tbl, cols, passthrough_deferred=passthrough_deferred)
                ]

        if unique:
            self.make_unique()

    def __iter__(self):
        return iter(self.cols)

    def __add__(self, other):
        if isinstance(other, VinylColumn):
            if self.tbl != other.tbl:
                raise ValueError(
                    "Can only add a VinylColumn to a VinylColumnList from the same VinylTable"
                )
            return VinylColumnList(self.tbl, self.cols.append(other))
        elif not isinstance(other, VinylColumnList):
            raise ValueError(
                f"Can only add a VinylColumnList to another VinylColumnList, not a {type(other)}"
            )

        elif self.tbl.__hash__() != other.tbl.__hash__():
            raise ValueError(
                "Can only add a VinylColumnList to another VinylColumnList from the same VinylTable"
            )

        return VinylColumnList(
            self.tbl,
            [c.col for c in self.cols] + [c.col for c in other.cols],
            unique=self.unique and other.unique,
        )

    def __radd__(self, other):
        if isinstance(other, VinylColumnList):
            return other.__add__(self)

        return self.__add__(other)

    def windowize(self, window_, adj_object=False):
        # ensure col can be windowed
        new_obj = self if adj_object else copy.deepcopy(self)
        for col in new_obj.cols:
            try:
                windowed = col.col.over(window_)
                if isinstance(windowed, Deferred):
                    windowed = windowed.resolve(self.tbl)
                col.col = windowed
            except com.IbisTypeError:
                pass

        return new_obj

    def make_unique(self):
        names = []
        unique_cols = []
        zipped = list(zip(self.names, self.queryable))[
            ::-1
        ]  # reverse so latest iteration of column is kept
        for name, col in zipped:
            if name not in names:
                names.append(name)
                unique_cols.append(col)

        unique_cols = unique_cols[::-1]
        self.cols = [VinylColumn(self.tbl, col) for col in unique_cols]
        self.unique = True

    def reset_tbl(self, tbl):
        self.tbl = tbl

    @property
    def queryable(self):
        return [col.col for col in self.cols if col.col is not None]

    @property
    def lambdaized(self):
        return [col.lambdaized for col in self.cols if col.col is not None]

    @property
    def sources_as_deferred(self):
        return [getattr(_, col) for col in self.get_direct_col_sources(unique=True)]

    def sources_as_deferred_resolved(self, tbl):
        return [getattr(tbl, col) for col in self.get_direct_col_sources(unique=True)]

    @property
    def names(self):
        return [col.name for col in self.cols]

    @property
    def types(self):
        return [col.type for col in self.cols]

    @property
    def names_as_deferred(self):
        return [col.name_as_deferred for col in self.cols]

    def names_as_deferred_resolved(self, tbl):
        return [col.name_as_deferred_resolved(tbl) for col in self.cols]

    def get_direct_col_sources(self, unique=True):
        # ibis table, not vinylTable
        col_sources = []
        for col in self.queryable:
            if isinstance(col, Deferred):
                col = col.resolve(self.tbl)
            if callable(col):
                col = col(self.tbl)
            sel_sources = [i.name for i in col.op().find_topmost(ops.Field)]
            if unique:
                col_sources.extend(sel_sources)
            else:
                col_sources.append(sel_sources)

        if unique:
            col_sources = list(set(col_sources))

        return col_sources


class VinylSortColumnList(VinylColumnList):
    def __init__(
        self, tbl, cols, reverse=False, unique=False, passthrough_deferred=False
    ):
        super().__init__(tbl, cols, unique, passthrough_deferred=passthrough_deferred)
        self.reverse = reverse

    @property
    def sorted(self):
        final = []
        for so in self.cols:
            # columns must be resolved for this to work
            if isinstance(so.col, Deferred):
                so.col = so.col.resolve(self.tbl)
            op_it = so.col.op()
            if isinstance(op_it, ops.SortKey):
                # in this case ops_it is a tuple, where the first key is the column and the second is the sort order bool (true is ascending)
                col_expr = op_it.args[0].to_expr()
                out_it = (
                    col_expr.desc() if op_it.args[1] ^ self.reverse else col_expr.asc()
                )
            else:
                out_it = so.col.desc() if self.reverse else so.col.asc()
            final.append(out_it)

        return final

    @property
    def unsorted(self):
        final = []
        for so in self.cols:
            op_it = so.col.op()
            if isinstance(op_it, ops.SortKey):
                # in this case ops_it is a tuple, where the first key is the column and the second is the sort order bool (true is ascending)
                col_expr = op_it.args[0].to_expr()

            else:
                col_expr = so.col
                # in this case, asc is the default
            final.append(col_expr)

        return final

    @property
    def names_as_deferred_sorted(self):
        directions = self.sort_directions
        return [
            name_def.asc() if directions[i] else name_def.desc()
            for i, name_def in enumerate(self.names_as_deferred)
        ]

    def names_as_deferred_resolved_sorted(self, tbl):
        directions = self.sort_directions
        return [
            name_def.asc() if directions[i] else name_def.desc()
            for i, name_def in enumerate(self.names_as_deferred_resolved(tbl))
        ]

    @property
    def sort_directions(self):
        sort_directions = []
        for so in self.cols:
            op_it = so.col.op()
            if isinstance(op_it, ops.SortKey):
                sort_directions.append(op_it.args[1])
            else:
                sort_directions.append(True)

        return sort_directions



from __future__ import annotations

import re
from typing import Any, Literal

import ibis.expr.operations as ops
import ibis.expr.types as ir
import networkx as nx
import rustworkx as rx
from ibis.expr.datatypes import DataType

from vinyl.lib.table import VinylTable
from vinyl.lib.utils.graph import rustworkx_to_networkx


class Field:
    _relations: rx.PyDiGraph = rx.PyDiGraph()
    _relations_node_dict: dict[str, int] = {}
    _field_relations: rx.PyDiGraph = rx.PyDiGraph()
    _field_relations_node_dict: dict[str, int] = {}
    _source_class_dict: dict[ir.Table, Any] = {}
    parent_name: str | None
    parent_table: object | None
    name: str | None
    type: DataType | None
    description: str | None
    primary_key: bool
    unique_key: bool
    foreign_key: Field | None
    pii: bool

    def __init__(
        self,
        parent_name: str | None = None,
        parent_table: object | None = None,
        name: str | None = None,
        type: DataType | None = None,
        description: str | None = None,
        primary_key: bool = False,
        unique_key: bool = False,
        foreign_key: Field | None = None,
        pii: bool = False,
    ):
        self.parent_name = parent_name
        self.parent_table = parent_table
        self.name = name
        self.type = type
        self.description = description
        self.primary_key = primary_key
        self.unique_key = unique_key
        self.foreign_key = foreign_key
        self.pii = pii

    def update_source_class(self):
        self._source_class_dict[self.parent_table.tbl] = (
            self.parent_table._source_class
            if hasattr(self.parent_table, "_source_class")
            else None
        )

    def store_relations(self):
        self.add_node(self.parent_table, "relations")
        self.add_node(getattr(self.parent_table, self.name), "field")
        if self.foreign_key is not None:
            # note for below, foriegn key class has been imported, so self.foreign_key is an ibis expression at this point
            foreign_key_table = VinylTable(
                self.foreign_key.op().find(ops.UnboundTable)[0]
            )
            foreign_key_table._source_class = self._source_class_dict.get(
                foreign_key_table.tbl
            )
            foreign_key_table._is_vinyl_source = True
            self.add_edge(
                self.foreign_key.op().find(ops.UnboundTable)[0].to_expr(),
                self.parent_table,
                "relations",
            )
            self.add_edge(
                self.foreign_key, getattr(self.parent_table, self.name), "field"
            )

    def add_node(self, node: ir.Expr, type=Literal["field", "relations"]):
        hashed = str(hash(node._arg))
        if type == "field" and hashed not in self._field_relations_node_dict:
            self._field_relations_node_dict[hashed] = self._field_relations.add_node(
                node
            )
        elif type == "relations" and hashed not in self._relations_node_dict:
            self._relations_node_dict[hashed] = self._relations.add_node(node)

    def add_edge(
        self, node1: ir.Expr, node2: ir.Expr, type=Literal["field", "relations"]
    ):
        for node in [node1, node2]:
            if type == "field":
                self.add_node(node, type)
            elif type == "relations":
                self.add_node(node, type)

        if type == "field":
            self._field_relations.add_edge(
                self._field_relations_node_dict[str(hash(node1._arg))],
                self._field_relations_node_dict[str(hash(node2._arg))],
                1,
            )
        elif type == "relations":
            self._relations.add_edge(
                self._relations_node_dict[str(hash(node1._arg))],
                self._relations_node_dict[str(hash(node._arg))],
                1,
            )

    def asdict(self):
        out = {}
        for k, v in self.__dict__.items():
            if v is not None and k != "_relations" and k != "_field_relations":
                out[k] = v

        return out

    def update(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def export_relations_to_networkx(
        cls, shorten_name: bool = True, filter: list[str] | None = None
    ):
        G = rustworkx_to_networkx(cls._relations)
        # make sure node object is passed as data rather than name to prevent issues with renderer
        nx.set_node_attributes(G, {node: node for node in G.nodes()}, "node")

        if shorten_name or filter:
            nx.relabel_nodes(
                G, {node: node.get_name() for node in G.nodes()}, copy=False
            )

        if not filter:
            return G
        else:
            all_nodes = list(G.nodes())
            nodes_in_subgraph: set[str] = set()
            for fil in filter:
                regex = re.compile(fil)
                nodes_to_add = set(node for node in all_nodes if regex.search(node))
                nodes_in_subgraph = nodes_in_subgraph.union(nodes_to_add)

            return G.subgraph(nodes_in_subgraph)



import inspect


# Dynamically wrap inherited methods by default to ensure output is of same class rather than parent class
# Note: ensure_output must be defined in the class
def wrap_inherited_methods(wrapper_func_str="ensure_output", **wrapper_args):
    def class_decorator(cls):
        # Retrieve the wrapper function from the class (or its module)
        wrapper_func = (
            getattr(cls, wrapper_func_str, None) or globals()[wrapper_func_str]
        )

        for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
            # Check if the method is defined in the parent class and not in the current class
            if method.__qualname__.split(".")[0] != cls.__name__:
                # Apply the wrapper with arguments, then apply the resulting decorator to the method
                decorated_method = wrapper_func(**wrapper_args)(method)
                setattr(cls, name, decorated_method)
        return cls  # Return the modified class

    return class_decorator


def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False



import contextlib
from typing import Any, Callable

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import networkx as nx
import rustworkx as rx
from rich.markup import escape
from rich.text import Text

from vinyl.lib.utils.graphics import adjust_type


def rustworkx_to_networkx(
    graph: rx.PyDiGraph,
    pre_convert: Callable[..., Any] | None = None,
    post_convert: Callable[..., Any] | None = None,
) -> nx.DiGraph:
    """Converts a rustworkx graph to a networkx graph."""
    g = nx.DiGraph()
    cur_nodes = graph.nodes()
    if pre_convert is not None:
        cur_nodes = [pre_convert(i) for i in cur_nodes]
    cur_edges = [(cur_nodes[u], cur_nodes[v]) for u, v in graph.edge_list()]
    g.add_nodes_from(nodes_for_adding=cur_nodes)
    g.add_edges_from(cur_edges)

    if post_convert is not None:
        nx.relabel_nodes(g, post_convert, copy=False)
    return g


## NOTE: Made an attempt at a fix here, but need to verify. Necessary because ops.Join is no longer a valid class
def get_type(node):
    with contextlib.suppress(AttributeError, NotImplementedError):
        return adjust_type(node.dtype)

    try:
        schema = node.schema
    except (AttributeError, NotImplementedError):
        # TODO(kszucs): this branch should be removed
        try:
            # As a last resort try get the name of the output_type class
            return node.output_type.__name__
        except (AttributeError, NotImplementedError):
            return "\u2205"  # empty set character
    except com.IbisError:
        assert isinstance(node, ops.JoinChain)
        tables = [node.first.parent]
        tables.extend([n.table.parent for n in node.rest])
        table_names = [getattr(t, "name", None) for t in tables]
        schemas = [t.schema for t in tables]
        pairs = []
        for i, table in enumerate(tables):
            pairs.extend(
                (f"{table_names[i]}.{column}", type)
                for column, type in schemas[i].items()
            )
        schema = ibis.schema(pairs)

    out = Text("\n   ")
    len_loop = len(schema.names)
    for i, (name, type) in enumerate(zip(schema.names, schema.types)):
        out += Text(escape(name), style="italic") + Text(": ") + adjust_type(type)
        if i < len_loop - 1:
            out += Text("\n   ")

    return out


def get_label(node):
    typename = get_type(node)  # Already an escaped string
    name = type(node).__name__
    nodename = (
        node.name
        if isinstance(
            node,
            (
                ops.Literal,
                ops.Field,
                ops.Alias,
                ops.PhysicalTable,
                ops.window.RangeWindowFrame,
            ),
        )
        else None
    )
    if nodename is not None:
        # [TODO] Don't show nodename because it's too long and ruins the image
        if isinstance(node, ops.window.RangeWindowFrame):
            label = Text(escape(name), style="bold")
        else:
            label = (
                Text(escape(nodename), style="italic")
                + Text(": ")
                + Text(escape(name), style="bold")
            )

            if isinstance(node, ops.Relation):
                label += typename
            else:
                label += Text("\n   :: ") + typename

    else:
        label = Text(escape(name), style="bold")
        if isinstance(node, ops.Relation):
            label += typename
        else:
            label += Text("\n   :: ") + typename

    return label



import os
from typing import Any, Callable

import ibis
from ibis.common.exceptions import IbisError

from vinyl.lib.connect import DatabaseFileConnector, _DatabaseConnector, _TableConnector
from vinyl.lib.constants import PreviewHelper
from vinyl.lib.field import Field
from vinyl.lib.project import load_project_module
from vinyl.lib.table import VinylTable


def _get_twin_relative_path(name: str, *subdirs) -> str:
    return os.path.join(
        "data",
        *subdirs,
        f"{name}.duckdb",
    )


def source(resource: Callable[..., Any], sample_row_count: int = 1000):
    def decorator(cls) -> VinylTable:
        connector = resource()

        schema = []
        for name, attr in cls.__annotations__.items():
            schema.append((name, attr))

        ibis.table(schema, name=cls._unique_name)
        table = VinylTable.create_from_schema(
            ibis.Schema.from_tuples(schema), cls._unique_name
        )
        parent_name = ".".join(cls.__module__.split(".")[2:])

        for name, attr in cls.__annotations__.items():
            field = Field(
                name=name, type=attr, parent_table=table, parent_name=parent_name
            )
            if hasattr(cls, name):
                current = getattr(cls, name)
                # Field info overrides the class level attributes, may want to change
                field.update(**current.asdict())
            setattr(cls, name, field)

        table = VinylTable.create_from_schema(
            ibis.Schema.from_tuples(schema), cls._unique_name
        )

        table._is_vinyl_source = True  # helps find sources in the project for load_defs
        table._source_class = cls

        # update parent table to include the full set of annotations and save graph
        for name, attr in cls.__annotations__.items():
            current = getattr(cls, name)
            # Field info overrides the class level attributes, may want to change
            current.update(**{"parent_table": table})
            setattr(cls, name, current)
            current.store_relations()

        unbounded_op = table.tbl.op()
        if isinstance(connector, _DatabaseConnector):

            def full_connection():
                return connector.get_table(
                    database=cls._database, schema=cls._schema, table=cls._table
                ).op()

            if PreviewHelper.preview == "full":
                table._conn_replace = {unbounded_op: full_connection}
            elif PreviewHelper.preview == "twin":
                # get twin
                twin_conn = DatabaseFileConnector(
                    os.path.join(
                        os.path.dirname(load_project_module().__file__),
                        _get_twin_relative_path(resource.__name__),
                    )
                )

                try:
                    table._twin_conn_replace = {
                        unbounded_op: lambda: twin_conn.get_table(
                            database=twin_conn.database,
                            schema=cls._schema,
                            table=cls._table,
                        ).op()
                    }
                except IbisError:
                    # twin doesn't exist, fall back to full conn, but not connected yet
                    table._conn_replace = {unbounded_op: full_connection}

            else:
                raise ValueError(
                    f"Invalid value for PreviewHelper.preview: {PreviewHelper.preview}"
                )
        elif isinstance(connector, _TableConnector):

            def full_connection():
                # for file connectors, we reconnect to the individual file to get the correct table. Since these tables are not in memory, we need to read to get the location.
                adj_conn = type(connector)(cls._path)
                adj_conn.connect()
                tbl_for_op = next(iter(adj_conn.tbls.values()))
                return tbl_for_op.op()

            if PreviewHelper.preview == "full":
                table._conn_replace = {unbounded_op: full_connection}

            elif PreviewHelper.preview == "twin":
                try:
                    table._twin_conn_replace = {
                        unbounded_op: table.tbl.filter(
                            ibis.random()
                            < ibis.least(sample_row_count / cls._row_count, 1)
                        ).op()
                    }
                except IbisError:
                    # twin doesn't exist, fall back to full conn, but not connected yet
                    table._conn_replace = {unbounded_op: full_connection}
        else:
            raise NotImplementedError(
                f"Connector type {type(connector)} is not yet supported"
            )

        return table

    return decorator



import os
from functools import wraps

from pydantic import validate_call

validate = validate_call(
    validate_return=False, config={"arbitrary_types_allowed": True}
)


def with_modified_env(var_name, new_value):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Save the original value of the environment variable (if it exists)
            original_value = os.environ.get(var_name, None)

            # Set the new value for the environment variable
            os.environ[var_name] = new_value

            try:
                # Call the actual function
                result = func(*args, **kwargs)
            finally:
                # Restore the original value of the environment variable
                if original_value is not None:
                    os.environ[var_name] = original_value
                else:
                    # The environment variable was not set before, so remove it
                    del os.environ[var_name]

            return result

        return wrapper

    return decorator



from sqlglot import exp
from sqlglot.dialects import DuckDB, Postgres, Trino
from sqlglot.tokens import TokenType


# Additional expressions I need
class ToISO8601(exp.Func):
    pass


class BVPostgres(Postgres):
    pass


# Trino-specific modifications
class BVTrino(Trino):
    class Tokenizer(Trino.Tokenizer):
        KEYWORDS = {
            **Trino.Tokenizer.KEYWORDS,
            "DEALLOCATE": TokenType.COMMAND,
        }

    class Parser(Trino.Parser):
        FUNCTIONS = {
            **Trino.Parser.FUNCTIONS,
            "TO_ISO8601": ToISO8601.from_arg_list,
        }


# DuckDB-specific modifications
def _duckdb_command_handler(self, expression):
    cmd = expression.this.upper()
    if cmd == "PREPARE":
        literal = expression.expression.this
        tokens = literal.split()
        stmt = tokens[0]
        if tokens[1].upper() not in ("FROM", "AS"):
            raise Exception("Badness badness")
        res = " ".join(tokens[2:])
        rest = BVDuckDB().generate(BVTrino().parse(res)[0])
        return f"PREPARE {stmt} AS " + rest
    elif cmd == "SHOW":
        literal = expression.expression.this
        tokens = literal.split()
        entity = tokens[0].upper()
        if entity == "CATALOGS":
            q = "SELECT DISTINCT catalog_name as Catalog FROM information_schema.schemata"
            if len(tokens) >= 3:
                if len(tokens) == 5:
                    like = tokens[2].replace(tokens[4], "")
                else:
                    like = tokens[2]
                q += " WHERE catalog_name LIKE " + like
            else:
                return q
        elif entity == "SCHEMAS":
            q = "SELECT DISTINCT schema_name as Schema FROM information_schema.schemata"
            if len(tokens) > 1 and tokens[1].upper() == "FROM":
                q += f" WHERE catalog_name = '{tokens[2]}'"
                if len(tokens) >= 5:
                    if len(tokens) == 7:
                        like = tokens[4].replace(tokens[6], "")
                    else:
                        like = tokens[4]
                    q += " AND schema_name LIKE " + like
            else:
                q += " WHERE catalog_name IN (SELECT current_database())"
                if len(tokens) >= 3:
                    if len(tokens) == 5:
                        like = tokens[2].replace(tokens[4], "")
                    else:
                        like = tokens[2]
                    q += " AND schema_name LIKE " + like
            return q
        elif entity == "TABLES":
            q = "SELECT DISTINCT table_name as Table from information_schema.tables"
            if len(tokens) > 1 and tokens[1].upper() == "FROM":
                q += f" WHERE table_schema = '{tokens[2]}'"
                if len(tokens) >= 5:
                    if len(tokens) == 7:
                        like = tokens[4].replace(tokens[6], "")
                    else:
                        like = tokens[4]
                    q += " AND table_name LIKE " + like
            else:
                q += " WHERE table_schema IN (SELECT current_schema())"
                if len(tokens) >= 3:
                    if len(tokens) == 5:
                        like = tokens[2].replace(tokens[4], "")
                    else:
                        like = tokens[2]
                    q += " AND table_name LIKE " + like
            return q
        elif entity == "COLUMNS" and tokens[1].upper() == "FROM":
            return f"DESCRIBE {tokens[2]}"
        elif entity == "TRANSACTION":
            return "SELECT 'read committed' as transaction_isolation"
        elif entity == "STANDARD_CONFORMING_STRINGS":
            return "SELECT 'on' as standard_conforming_strings"
        else:
            raise Exception("Unhandled SHOW command: " + literal)

    return expression.sql()


class BVDuckDB(DuckDB):
    class Generator(DuckDB.Generator):
        TRANSFORMS = {
            **DuckDB.Generator.TRANSFORMS,
            exp.CurrentTimestamp: lambda *_: "CURRENT_TIMESTAMP",
            ToISO8601: lambda self,
            e: f"STRFTIME({self.sql(e, 'this')}, '%Y-%m-%dT%H:%M:%S.%f%z')",
            exp.Command: _duckdb_command_handler,
        }



import random
import re
import string
from urllib.parse import urlparse


def generate_random_ascii_string(length=30):
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def split_interval_string(s):
    for i, char in enumerate(s):
        if not char.isdigit():
            return int(s[:i]), s[i:]
    return int(s), ""


def extract_uri_scheme(path):
    result = urlparse(path)
    return result.scheme


def make_python_identifier(str_):
    return re.sub(r"\W|^(?=\d)", "_", str_)



import itertools
from pathlib import Path

import toml


def get_root():
    # Create a Path object for the start directory
    path = Path().resolve()

    # Traverse up through the parent directories
    for parent in itertools.chain([path], path.parents):
        pyproject_toml = parent / "pyproject.toml"
        if pyproject_toml.exists():
            return pyproject_toml.parent.absolute()

    return None


class PyProjectSettings:
    toml: dict

    def __init__(self, path: Path | None = None):
        if path is None:
            path = get_root()
        with open(path / "pyproject.toml") as f:
            self.toml = toml.load(f)

    def get_setting(self, key: str):
        return self.toml["tool"]["vinyl"].get(key, None)



from enum import Enum
from typing import Any, Literal

import ibis.expr.types as ir
from ibis.common.deferred import Deferred
from lets_plot import (
    LetsPlot,
    aes,
    coord_flip,
    facet_grid,
    facet_wrap,
    flavor_darcula,
    ggplot,
    scale_x_datetime,
    scale_y_datetime,
)

from vinyl.lib.column import VinylColumn, VinylColumnList
from vinyl.lib.settings import PyProjectSettings


class geom(Enum):
    from lets_plot import (
        geom_area,
        geom_area_ridges,
        geom_bar,
        geom_bin2d,
        geom_boxplot,
        geom_density,
        geom_histogram,
        geom_line,
        geom_point,
        geom_smooth,
        geom_violin,
        position_dodge,
        ylab,
    )

    scatter = geom_point()
    line = geom_line()
    bar = geom_bar(stat="identity", position=position_dodge())
    area = geom_area()
    stacked_bar = geom_bar(stat="identity")
    percent_bar = geom_bar() + aes(y="..prop..") + ylab("Percent of total")
    histogram = geom_histogram()
    histogram_2d = geom_bin2d()
    violin = geom_violin()
    boxplot = geom_boxplot()
    density = geom_density()
    ridge = geom_area_ridges()
    trendline_lm = geom_smooth()
    trendline_loess = geom_smooth(method="loess")


class BaseChart:
    mode: Literal["light", "dark"] = "dark"

    from lets_plot import (
        LetsPlot,
        aes,
        coord_flip,
        facet_grid,
        facet_wrap,
        flavor_darcula,
        ggplot,
        scale_x_datetime,
        scale_y_datetime,
    )

    def __init__(
        self,
        geoms: geom | list[geom],
        source: Any,  # will be VinylTable, but use Any to avoid recursion
        x: Deferred | ir.Value | None,
        y: Deferred | ir.Value | None = None,
        color: Deferred | ir.Value | None = None,
        fill: Deferred | ir.Value | None = None,
        size: Deferred | ir.Value | None = None,
        alpha: Deferred | ir.Value | None = None,
        facet: Deferred | ir.Value | list[Deferred | ir.Value] | None = None,
        coord_flip: bool = False,
    ):
        self.geoms = geoms
        self.data = source
        self.x = x
        self.y = y
        self.color = color
        self.fill = fill
        self.size = size
        self.alpha = alpha
        self.facet = facet
        self.coord_flip = coord_flip

    def show(self):
        LetsPlot.setup_html()
        adj_facet = self.facet if isinstance(self.facet, list) else [self.facet]
        all_cols = [
            x
            for x in [self.x]
            + [self.y]
            + [self.color]
            + [self.fill]
            + [self.size]
            + [self.alpha]
            + adj_facet
            if x is not None
        ]

        adj_data = self.data.mutate(all_cols).execute("pandas")

        ## make sure all cols are in there,
        vinyl_x = VinylColumn(self.data.tbl, self.x)
        type_x = vinyl_x.type
        if self.y is not None:
            vinyl_y = VinylColumn(self.data.tbl, self.y)
            type_y = vinyl_y.type
        aes_dict = {}
        for var in ["x", "y", "color", "fill", "size"]:
            attr = getattr(self, var)
            if attr is not None:
                aes_dict[var] = VinylColumn(self.data.tbl, attr).name

        plot = ggplot(adj_data, aes(**aes_dict))
        if isinstance(self.geoms, list):
            for g in self.geoms:
                plot += g.value
        elif self.geoms is not None:
            plot += self.geoms.value
        if self.facet is not None:
            facet_names = VinylColumnList(self.data.tbl, adj_facet).names
            if len(adj_facet) > 1:
                plot += facet_grid(
                    facet_names[0],
                    facet_names[1],
                )
            else:
                plot += facet_wrap(facet_names[0])
        if type_x.is_timestamp():
            plot += scale_x_datetime()
        if self.y is not None and type_y.is_timestamp():
            plot += scale_y_datetime()
        if self.coord_flip:
            plot += coord_flip()

        if PyProjectSettings().get_setting("dark-mode") is True:
            plot += flavor_darcula()

        return plot



from __future__ import annotations

from typing import Any, Callable, Literal, Sequence, TypeAlias

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis import selectors as s
from ibis.common.deferred import Deferred

from vinyl.lib.column import VinylColumnList, base_column_type, column_type_all
from vinyl.lib.enums import FillOptions
from vinyl.lib.utils.text import generate_random_ascii_string

fill_type: TypeAlias = (
    list[FillOptions | Callable[..., Any] | None]
    | FillOptions
    | Callable[..., Any]
    | None
)


def expand_col(
    tbl: ir.Table, col: ir.Value, adjust_name: bool = False
) -> tuple[ir.Table, dict[str, str] | None]:
    vinyl_col = VinylColumnList(tbl, col)
    cur_col_name = vinyl_col.names[0]
    type_ = vinyl_col.types[0]
    col_key = "col_it"
    col_tbl = tbl.select(vinyl_col.queryable)
    default_return = col_tbl.distinct()

    if not type_.is_timestamp() and not type_.is_date():
        return default_return, None

    if isinstance(col, Deferred):
        col = col.resolve(tbl)
    col_op = col.op()
    last_fn_list = col_op.find_topmost(ops.TimestampBucket) + col_op.find_topmost(
        ops.TimestampTruncate
    )
    last_fn = last_fn_list[0] if len(last_fn_list) > 0 else None
    original_col = VinylColumnList(tbl, col_op.find_topmost(ops.Field)[0].to_expr())
    new_col_name = original_col.names[0] if adjust_name else None

    if last_fn is None:
        return default_return, None

    temp_renamed_col_tbl = col_tbl.rename({col_key: cur_col_name})

    range_helper = temp_renamed_col_tbl.aggregate(
        metrics=[
            getattr(_, col_key).min().name("min"),
            getattr(_, col_key).max().name("max"),
        ]
    )
    sel_name = new_col_name if new_col_name is not None else cur_col_name
    sel_name_for_rename_dict = (
        {cur_col_name: new_col_name} if new_col_name is not None else {}
    )

    if type_.is_timestamp():
        if isinstance(last_fn, ops.TimestampTruncate):
            range_ = (
                ops.TimestampRange(
                    start=range_helper.min,
                    stop=range_helper.max
                    + ibis.interval(
                        unit=last_fn.unit, value=1
                    ),  # need inclusive range, ibis stop is exclusive by default
                    step=ibis.interval(unit=last_fn.unit, value=1),
                )
                .to_expr()
                .name(sel_name)
                .as_table()
                .select(_[sel_name].unnest())
            )
            return ir.Table(range_._arg), sel_name_for_rename_dict

        elif isinstance(last_fn, ops.TimestampBucket):
            range_ = (
                ops.TimestampRange(
                    start=range_helper.min,
                    stop=range_helper.max
                    + last_fn.interval,  # need inclusive range, ibis stop is exclusive by default
                    step=last_fn.interval,
                )
                .to_expr()
                .name(sel_name)
                .as_table()
                .select(_[sel_name].unnest())
            )
            return ir.Table(range_._arg), sel_name_for_rename_dict

        else:
            return default_return, None

    return default_return, None


def _auto_join(tbl: ir.Table, tbl2: ir.Table):
    common_columns = list(set(tbl.schema().names) & set(tbl2.schema().names))
    if len(common_columns) == 0:
        return tbl.cross_join(tbl2)
    else:
        return tbl.outer_join(tbl2, predicates=common_columns)


def _build_spine(
    tbl: ir.Table, cols: Sequence[ir.Scalar]
) -> tuple[ir.Table, dict[str, Any] | None]:
    # no need to go through expansion logic if there is only one column and there's no interpolation
    vinyl_cols = VinylColumnList(tbl, cols)
    if len(vinyl_cols.cols) == 1:
        return expand_col(tbl, vinyl_cols.queryable[0])

    rename_dict = {}
    zipped = list(zip(vinyl_cols.names, vinyl_cols.queryable))
    for i, (name, col) in enumerate(zipped):
        expanded_col_tpl = expand_col(tbl, col)
        if expanded_col_tpl[1]:
            rename_dict.update(expanded_col_tpl[1])

        if i == 0:
            spine = expanded_col_tpl[0]
        else:
            spine = _auto_join(spine, expanded_col_tpl[0])

    return spine, rename_dict


def _join_spine(self, spine: ir.Table) -> ir.Table:
    # generate random col_name to remove extra join columns (ibis includes all cols from left and right by default)
    random_col_suffix = generate_random_ascii_string()
    rname = f"{{name}}_{random_col_suffix}"

    null_fill = spine.left_join(  # ensure using ibis join
        self,
        predicates=[(spine[n], self[n]) for n in spine.schema().names],
        rname=rname,
    ).drop(s.contains(random_col_suffix))  # perform the drop

    return null_fill


def _adjust_fill_list(
    n_col: int,
    fill: list[FillOptions | Callable[..., Any] | None]
    | FillOptions
    | Callable[..., Any]
    | None = None,
):
    fill_list: list[FillOptions | Callable[..., Any] | None]
    if isinstance(fill, FillOptions):
        fill_list = [fill] * n_col
    elif isinstance(fill, list):
        fill_list = fill
    else:
        # is a lambda
        fill_list = [fill] * n_col

    if len(fill_list) != n_col:
        raise ValueError(
            "Must provide either a single fill option or a fill option for each aggregation"
        )

    return fill_list


def _union_adjuster(
    first: ir.Table, second: ir.Table, distinct: bool = False
) -> ir.Table:
    # get current schemas
    first_schema = first.schema()
    first_names = first_schema.names
    second_schema = second.schema()
    second_names = second_schema.names

    # check if any common cols have different types, raise error if they do
    common_cols = set(first_names) & set(second_names)
    common_cols_with_diff_types = [
        col for col in common_cols if first_schema[col] != second_schema[col]
    ]
    if len(common_cols_with_diff_types) > 0:
        raise ValueError(
            f"Cannot union tables with common columns of different types: {common_cols_with_diff_types}"
        )
    # get correct column order (self + other not in self)
    all_cols_with_duplicates = first_names + second_names
    all_cols = list(set(all_cols_with_duplicates))

    # if no columns to add, return union
    if (
        len(all_cols_with_duplicates)
        == len(first_schema.names)
        == len(all_cols_with_duplicates)
    ):
        return first.union(second, distinct=distinct)

    # get columns to add
    self_cols_to_add = [
        ibis.NA.cast(second_schema[col]).name(col)
        for col in all_cols
        if col not in first_schema.names
    ]
    other_cols_to_add = [
        ibis.NA.cast(first_schema[col]).name(col)
        for col in all_cols
        if col not in second_schema.names
    ]

    # add missing columns to each table
    adjusted_first_cols = [getattr(_, name) for name in first_names] + self_cols_to_add
    adjusted_first = (
        first.select(*adjusted_first_cols) if len(self_cols_to_add) > 0 else first
    )
    adjusted_second_cols = [
        getattr(_, name) for name in second_names
    ] + other_cols_to_add
    adjusted_second = (
        second.select(*adjusted_second_cols) if len(other_cols_to_add) > 0 else second
    )

    return adjusted_first, adjusted_second


def _union_two(first: ir.Table, second: ir.Table, distinct: bool = False) -> ir.Table:
    adjusted_first, adjusted_second = _union_adjuster(first, second, distinct)
    return adjusted_first.union(adjusted_second, distinct=distinct)


def _difference_two(
    first: ir.Table, second: ir.Table, distinct: bool = False
) -> ir.Table:
    adjusted_first, adjusted_second = _union_adjuster(first, second, distinct)
    return adjusted_first.difference(adjusted_second, distinct=distinct)


def _join_with_removal(
    self: ir.Table,
    other: ir.Table,
    how: Literal[
        "inner",
        "left",
        "outer",
        "right",
        "semi",
        "anti",
        "any_inner",
        "any_left",
        "left_semi",
    ] = "outer",
) -> ir.Table:
    random_col_suffix = generate_random_ascii_string()
    rname = f"{{name}}_{random_col_suffix}"
    common_columns = list(set(self.columns) & set(other.columns))
    if len(common_columns) == 0:
        return self.cross_join(other)
    else:
        return self.join(other, predicates=common_columns, rname=rname, how=how).drop(
            s.contains(random_col_suffix)
        )


def _select_process_helper(tbl: ir.Table, se: base_column_type):
    name_dict = {}
    vinyl_se = VinylColumnList(tbl, se)
    if isinstance(se, s.Selector):
        name_dict = {col.name: col.col for col in vinyl_se.cols}

    else:
        cur_name = vinyl_se.names[0]
        sources = vinyl_se.get_direct_col_sources(unique=False)[0]
        adj_name = sources[0] if sources != [] else cur_name
        name_dict[adj_name] = se

    return name_dict


def _process_multiple_select(
    tbl: ir.Table,
    col_selector: column_type_all,
    f: Callable[[Any], Any] | list[Callable[[Any], Any] | None] | None,
    rename=True,
):
    col_selector_list = (
        [col_selector] if not isinstance(col_selector, list) else col_selector
    )
    f_list: list[Callable[..., Any] | None] = [f] if not isinstance(f, list) else f

    if len(f_list) == 1 and len(col_selector_list) > 1:
        f_list = f_list * len(col_selector_list)

    if len(col_selector_list) != len(f_list):
        raise ValueError(
            "Must provide either a single function or a function for each column or selector"
        )

    adj_cols = []
    # make sure col_selector is enumerable
    if isinstance(col_selector, (str, s.Selector, ir.Value)) or callable(col_selector):
        col_selector = [col_selector]
    for i, sel in enumerate(col_selector):
        name_dict = {}
        if isinstance(sel, list):
            for se in sel:
                name_dict.update(_select_process_helper(tbl, se))
        else:
            name_dict.update(_select_process_helper(tbl, sel))

        for name, col in name_dict.items():
            f_it = f_list[i]  # extra step to help mypy
            if f_it is not None:
                core = f_it(col)
            else:
                core = col
            core = core.name(name) if not rename else core
            adj_cols.append(core)

    return adj_cols



import os

from vinyl.lib.asset import resource
from vinyl.lib.connect import FileConnector


@resource
def local_filesystem():
    return FileConnector(path=os.path.join(os.path.dirname(__file__), "data"))



import enum
import json
import re
import uuid
from typing import Any, Dict, Iterator, List, Optional, Tuple


class BVType(enum.Enum):
    NULL = 0
    BIGINT = 1
    BOOL = 2
    BYTES = 3
    DATE = 4
    FLOAT = 5
    INTEGER = 6
    INTERVAL = 7
    JSON = 8
    DECIMAL = 9
    TEXT = 10
    TIME = 11
    TIMESTAMP = 12
    UNKNOWN = 13
    ARRAY = 14
    INTEGERARRAY = 15
    STRINGARRAY = 16


class QueryResult:
    """The BV representation of a result of a query."""

    def __init__(self):
        self.result_format = None

    def has_results(self) -> bool:
        raise NotImplementedError

    def column_count(self):
        raise NotImplementedError

    def column(self, index: int) -> Tuple[str, BVType]:
        raise NotImplementedError

    def rows(self) -> Iterator[List]:
        raise NotImplementedError

    def status(self) -> str:
        raise NotImplementedError


class Session:
    def __init__(self):
        self.id = uuid.uuid4()

    def cursor(self):
        raise NotImplementedError

    def close(self):
        raise NotImplementedError

    def execute_sql(self, sql: str, params=None) -> QueryResult:
        raise NotImplementedError

    def in_transaction(self) -> bool:
        raise NotImplementedError

    def load_df_function(self, table: str):
        raise NotImplementedError


class Connection:
    """Translation layer from an upstream data source into the BV representation of a query result."""

    def __init__(self):
        self._sessions = {}

    def create_session(self) -> Session:
        sess = self.new_session()
        self._sessions[sess.id] = sess
        return sess

    def get_session(self, id: int) -> Optional[Session]:
        return self._sessions.get(id)

    def close_session(self, session: Session):
        if session and session.id in self._sessions:
            del self._sessions[session.id]
            session.close()

    def new_session(self) -> Session:
        raise NotImplementedError

    def parameters(self) -> Dict[str, str]:
        return {}


class Extension:
    @classmethod
    def check_json(cls, payload: str) -> Optional[dict]:
        is_json = False
        if payload[-1] == "}":
            is_json = True
        elif payload[-1] == ";" and payload[-2] == "}":
            is_json = True
        # Strip any SQL comments
        if is_json:
            payload = re.sub(r"\/\*.*\*\/", "", payload)
            return json.loads(payload)
        return None

    def type(self) -> str:
        raise NotImplementedError

    def apply(self, params: dict, session: Session) -> QueryResult:
        raise NotImplementedError


class SimpleQueryResult(QueryResult):
    def __init__(self, name: str, value: Any, type: BVType):
        super().__init__()
        self.name = name
        self.value = str(value)
        self.type = type

    def has_results(self):
        return True

    def column_count(self):
        return 1

    def column(self, index: int) -> Tuple[str, BVType]:
        if index == 0:
            return (self.name, self.type)
        else:
            raise IndexError

    def rows(self) -> Iterator[List]:
        return iter([[self.value]])

    def status(self) -> str:
        return ""



import dataclasses
import os
from importlib import import_module
from typing import Any

from vinyl.lib.project import get_project_module_name, load_project_module
from vinyl.lib.utils.pkg import (
    find_submodules_names,
)


@dataclasses.dataclass
class Defs:
    resources: list[Any]
    sources: list[Any]
    models: list[Any]
    metrics: list[Any]


def load_defs() -> Defs:
    project_module_name = get_project_module_name()

    modules = []
    modules.append(import_module(f"{project_module_name}.resources"))
    modules.append(import_module(f"{project_module_name}.sources"))
    no_models_vinyl = os.getenv("NO_MODELS_VINYL")
    if no_models_vinyl is None or no_models_vinyl == "False":
        modules.append(import_module(f"{project_module_name}.models"))

    defs_instance = Defs(resources=[], sources=[], models=[], metrics=[])

    for m in modules:
        for name in find_submodules_names(m):
            module = import_module(name)
            for func in dir(module):
                attr = getattr(module, func)
                if func != "_":
                    # order here is important, resources, but be imported before sources
                    for key in ["resource", "source", "model", "metric_store"]:
                        plural_key = key + "s"
                        if hasattr(attr, f"_is_vinyl_{key}"):
                            current = getattr(defs_instance, plural_key)
                            if attr not in current:
                                setattr(
                                    defs_instance,
                                    plural_key,
                                    getattr(defs_instance, plural_key) + [attr],
                                )
    return defs_instance


def load_project_defs() -> Defs:
    imported_module = load_project_module()
    return imported_module.defs
